import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from sklearn.mixture import GaussianMixture

## concatenate all matrices
path = "C:/Users/heyu/PycharmProjects/backup/big matrix records/"
oldBigMatrix = []
for i in range(4):  # 4 files
    filename = "bigMatrix_studyTestUsers_" + str(i) + ".txt"
    with open(path+filename,"r") as file:
        matrix = eval(file.readline())
        oldBigMatrix.extend(matrix)
print(np.array(oldBigMatrix).shape)

# remove duplicated sessions
bigMatrix = []
for i in oldBigMatrix:
    if(i not in bigMatrix):
        bigMatrix.append(i)
print(np.array(bigMatrix).shape)

visitors = len(bigMatrix)
sessions = len(bigMatrix[0])
print(visitors,sessions)

## use pca to reduce dimensions
pca_3d = []
for i in range(visitors):
    pca_list = []
    for j in range(sessions):
        userflow = bigMatrix[i][j]
        pca = PCA(n_components=1)
        principalComponents = pca.fit_transform(userflow)
        df = pd.DataFrame(data = principalComponents)
        df = df[0].values.tolist()
        
        pca_list.append(df)
    pca_3d.append(pca_list)  # till here every 33-list in pca_3d is corresponding to the userflow matrix

# apply pca again
pca1_list =[]
for m in range(visitors):
    pca1 = PCA(n_components=1)
    principalComponents1 = pca1.fit_transform(pca_3d[m])
    df1 = pd.DataFrame(data = principalComponents1)
    df1 = df1[0].values.tolist()
    pca1_list.append(df1)

# round numbers to 2 decimals. 
# the cluster results of 5 decimals, 3 decimals and before rounding are the same
# the heatmap is created by around 2 decimals
for p in range(len(pca1_list)):
    for q in range(len(pca1_list[p])):
        pca1_list[p][q] = round(pca1_list[p][q],2)
#print(pca1_list)
pca1_list = np.array(pca1_list)

## use bayesian information criterion to choose the number of clusters
X = pca1_list
num_centers = [2,3,4,5,6,7,8,9,10,11,12,13,14,15,16]
bics = []
for num in num_centers:
    gmm = GaussianMixture(n_components=num)
    gmm = gmm.fit(X)
    bic = gmm.bic(X)  # the lower the better
    bics.append(bic)
print(bics)
# plot bics
plt.figure()
plt.plot(num_centers,bics)
plt.xticks(np.arange(min(num_centers),max(num_centers)+1,step=1))
plt.xlabel("Number of centers")
plt.ylabel("Bayesian information criterion")
plt.title("User flow data BIC result")
plt.show()
# the numver of clusters
k = num_centers[np.argmin(bics,axis=0)]
print(k)

## apply GMM
gmm = GaussianMixture(n_components=15)
gmm = gmm.fit(X)
clusters = gmm.predict(X)
print(clusters)
probs = gmm.predict_proba(X)
#print(probs)
bic = gmm.bic(X)  # bayesian information criterion
print(bic)

## heatmap of all users
x_labels = [str('s'+ str(i)) for i in range(sessions)]
y_labels = [str('u'+ str(i)) for i in range(visitors)]
df_cm = pd.DataFrame(pca1_list,y_labels, x_labels)
plt.figure(figsize=(10,35))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-5,vmax=25)# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()

## heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == 14]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 14 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-4,vmax=16)# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()
