import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from sklearn.mixture import GaussianMixture
import time
from time import mktime
from datetime import timedelta
from itertools import groupby
import plotly
plotly.offline.init_notebook_mode()

## choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','amar','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get education list
path = 'C:/Users/heyu/PycharmProjects/education_list.csv'   
df = pd.read_csv(path)
educationId_list = list(df['0'].values)
print(len(educationId_list))

## get csv data and put it into a matrix with 3 features
path = 'C:/Users/heyu/PycharmProjects/content data'   
features_list = []
avg_num_education_list = []
education_freq_matrix = []
complete_ratio_list = []
for i in range(usersNum):
    # get completed data
    filename1 = '/' + usersName[i] + '_education_data.csv'
    df1 = pd.read_csv(path+filename1)
    educationIdCom = df1['educationid'].values
    timestampCom = df1['datecompleted'].values
    timestampCom.sort()
    tsCom_uni = []
    for j in range(len(timestampCom)):
        if(not(timestampCom[j] in tsCom_uni)):
            tsCom_uni.append(timestampCom[j])
    
    # get suggested data
    filename2 = '/' + usersName[i] + '_education_data_suggested.csv'
    df2 = pd.read_csv(path+filename2)
    educationIdSug = df2['educationid'].values
    timestampSug = df2['datesuggested'].values
    timestampSug.sort()
    tsSug_uni = []
    for j in(range(len(timestampSug))):
        if(not(timestampSug[j] in tsSug_uni)):
            tsSug_uni.append(timestampSug[j])
    
    # remove duplicated timestamp
    duplicate = 0
    for j in range(len(tsSug_uni)-1):
        if((tsSug_uni[j+1]-tsSug_uni[j])<1000):
            duplicate = tsSug_uni[j]
    df2 = df2[df2.datesuggested != duplicate]
    timestampSug = df2['datesuggested'].values
    timestampSug.sort()
    tsSug_uni = []
    for j in(range(len(timestampSug))):
        if(not(timestampSug[j] in tsSug_uni)):
            tsSug_uni.append(timestampSug[j])
    
    # use education ids in all sessions as feature 1, showing frequency information
    education_freq_list = [0]*len(educationId_list)
    for j in educationIdCom:
        index_j = educationId_list.index(j)
        education_freq_list[index_j] +=1
    education_freq_matrix.append(education_freq_list)
    
    # separate sessions
    sessionsCom = [list(g) for k, g in groupby(timestampCom)]
    sessionsSug = [list(g) for k, g in groupby(timestampSug)]

    # number of educations in each session
    sessionCom_num = []
    for j in range(len(sessionsCom)):
        sessionCom_num.append(len(sessionsCom[j]))
    sessionSug_num = []
    for j in range(len(sessionsSug)):
        sessionSug_num.append(len(sessionsSug[j]))
    
    # calculate average number of educations in one session (sessions with 0 education are not included) -- feature2
    sum_num_education = 0
    initial = 0
    for j in range(len(sessionCom_num)):
        sum_num_education = sum_num_education + sessionCom_num[j]
    avg_num_education = sum_num_education / len(sessionCom_num)
    avg_num_education_list.append(avg_num_education)
    
    # complete ratio of each week -- feature 3
    # calculate the expected timestamp
    tsExp = []
    for j in range(len(tsSug_uni)):
        ts_exp = tsSug_uni[j] + 604800000  # add 7 days
        tsExp.append(ts_exp)
    
    # get the suggested timestamp with educations completed in that week
    tsSugCom = []
    for j in range(len(tsCom_uni)):
        for k in range(len(tsExp)):
            if(abs(tsCom_uni[j]-tsExp[k])<86400000):  # smaller than 1 day
                tsSugCom.append([tsCom_uni[j],tsSug_uni[k]])  # put the completed timestamp and the corresponding suggested timestamp into a pair
    
    # complete_ratio 
    complete_ratio = [0]*len(sessionSug_num)
    if(len(sessionSug_num)==len(sessionCom_num)):  # if every week there are educations completed
        for j in range(len(sessionSug_num)):
            complete_ratio[j] = sessionCom_num[j] / sessionSug_num[j]  # for weeks with educations cpmpleted
    else:  # if there is a week or weeks with no education completed
        for j in tsSugCom:
            index_sug = tsSug_uni.index(j[1])
            index_com = tsCom_uni.index(j[0])
            complete_ratio[index_sug] = sessionCom_num[index_com] / sessionSug_num[index_sug]  # put non_zero complete ratio to the weeks with educations completed
    complete_ratio_list.append(complete_ratio)

# apply pca on education_freq_matrix
pca = PCA(n_components=1)
principalComponents0 = pca.fit_transform(education_freq_matrix)
education_freq_users = pd.DataFrame(data = principalComponents0)
education_freq_users = education_freq_users[0].values.tolist()

# pad 0s to complete ratio list to make matrix
pad = len(max(complete_ratio_list,key=len))
complete_ratio_list = np.array([i + [0]*(pad-len(i)) for i in complete_ratio_list])

# apply pca on complete_ratio_matrix
pca = PCA(n_components=1)
principalComponents1 = pca.fit_transform(complete_ratio_list)
complete_ratio_users = pd.DataFrame(data = principalComponents1)
complete_ratio_users = complete_ratio_users[0].values.tolist()

features_list.append(education_freq_users)
features_list.append(avg_num_education_list)
features_list.append(complete_ratio_users)

# transpose the features list
features_list = np.transpose(features_list)

## use bayesian information criterion to choose the number of clusters
X = features_list
num_centers = [2, 3, 4, 5, 6]
bics = []
for num in num_centers:
    gmm = GaussianMixture(n_components=num)
    gmm = gmm.fit(X)
    bic = gmm.bic(X)  # the lower the better
    bics.append(bic)
print(bics)
# plot bics
plt.figure()
plt.plot(num_centers,bics)
plt.xticks(np.arange(min(num_centers),max(num_centers)+1,step=1))
plt.xlabel("Number of centers")
plt.ylabel("Bayesian information criterion")
plt.title("Education data BIC result")
plt.show()
# the numver of clusters
k = num_centers[np.argmin(bics,axis=0)]
print(k)

## apply GMM
gmm = GaussianMixture(n_components=2)
gmm = gmm.fit(X)
clusters = gmm.predict(X)
print(clusters)
probs = gmm.predict_proba(X)
bic = gmm.bic(X)  # bayesian information criterion
print(bic)

## add cluster ids to features_list
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = np.transpose(clusters)

## plot clustering result with shape
cluster0 = df_cm.loc[df_cm['cluster'] == 0]
cluster1 = df_cm.loc[df_cm['cluster'] == 1]
cluster2 = df_cm.loc[df_cm['cluster'] == 2]
cluster3 = df_cm.loc[df_cm['cluster'] == 3]
cluster4 = df_cm.loc[df_cm['cluster'] == 4]
cluster5 = df_cm.loc[df_cm['cluster'] == 5]

scatter0 = dict(
    mode = "markers",
    name = "Cluster 0",
    type = "scatter3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    marker = dict( size=2, color='green')
)
scatter1 = dict(
    mode = "markers",
    name = "Cluster 1",
    type = "scatter3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    marker = dict( size=2, color='blue')
)
scatter2 = dict(
    mode = "markers",
    name = "Cluster 2",
    type = "scatter3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    marker = dict( size=2, color='red')
)
scatter3 = dict(
    mode = "markers",
    name = "Cluster 3",
    type = "scatter3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    marker = dict( size=2, color='magenta')
)
scatter4 = dict(
    mode = "markers",
    name = "Cluster 4",
    type = "scatter3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    marker = dict( size=2, color='yellow')
)
scatter5 = dict(
    mode = "markers",
    name = "Cluster 5",
    type = "scatter3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    marker = dict( size=2, color='grey')
)
cluster0 = dict(
    alphahull = 5,
    name = "Cluster 0",
    opacity = .1,
    type = "mesh3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    color='green', showscale = True
)
cluster1 = dict(
    alphahull = 5,
    name = "Cluster 1",
    opacity = .1,
    type = "mesh3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    color='blue', showscale = True
)
cluster2 = dict(
    alphahull = 5,
    name = "Cluster 2",
    opacity = .1,
    type = "mesh3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    color='red', showscale = True
)
cluster3 = dict(
    alphahull = 5,
    name = "Cluster 3",
    opacity = .1,
    type = "mesh3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    color='magenta', showscale = True
)
cluster4 = dict(
    alphahull = 5,
    name = "Cluster 4",
    opacity = .1,
    type = "mesh3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    color='yellow', showscale = True
)
cluster5 = dict(
    alphahull = 5,
    name = "Cluster 5",
    opacity = .1,
    type = "mesh3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    color='grey', showscale = True
)
layout = dict(
    title = 'Education data GMM clustering result ('+str(datetime.now().date())+")",
    scene = dict(
        xaxis = dict( title='edu_freq',zeroline=True ),
        yaxis = dict( title='avg_num_edu_per_session',zeroline=True ),
        zaxis = dict( title='complete_ratio',zeroline=True ),
    )
)
fig = dict( data=[scatter0, scatter1, scatter2, scatter3, scatter4, scatter5,cluster0, cluster1, cluster2,cluster3, cluster4, cluster5], layout=layout )
plotly.offline.iplot(fig)

## normalize features
edu_freq_df = pd.DataFrame(education_freq_users)
avg_num_edu_df = pd.DataFrame(avg_num_education_list)
complete_ratio_df = pd.DataFrame(complete_ratio_users)

edu_freq_norm = (edu_freq_df - edu_freq_df.mean()) / (edu_freq_df.max() - edu_freq_df.min())
avg_num_edu_norm = (avg_num_edu_df - avg_num_edu_df.mean()) / (avg_num_edu_df.max() - avg_num_edu_df.min())
complete_ratio_norm = (complete_ratio_df - complete_ratio_df.mean()) / (complete_ratio_df.max() - complete_ratio_df.min())

education_freq_users_norm = list(edu_freq_norm[0].values)
avg_num_education_list_norm = list(avg_num_edu_norm[0].values)
complete_ratio_users_norm = list(complete_ratio_norm[0].values)

features_list_norm = []
features_list_norm.append(education_freq_users_norm)
features_list_norm.append(avg_num_education_list_norm)
features_list_norm.append(complete_ratio_users_norm)

# round each number to 2 decimals
for i in range(len(features_list_norm)):
    for j in range(len(features_list_norm[i])):
        features_list_norm[i][j] = round(features_list_norm[i][j],2)

# transpose the features list
features_list = np.transpose(features_list_norm)

## heatmap of all users after normalization
x_labels = ['edu_freq','avg_num_edu_per_session','complete_ratio']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-0.4,vmax=0.6)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == 1]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 1 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-0.4,vmax=0.6)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
