import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
import json
import requests
from pandas.io.json import json_normalize
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import cophenet
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import fcluster
from collections import Counter
import matplotlib.patches as mpatches

## get backup sessions data
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "sessions" + ".txt"
sessionList = []
with open(path+filename,"r") as file:
    session = eval(file.readline())
    sessionList.extend(session)
    
## normalize time spent
flow_df = pd.DataFrame(sessionList[0])
time_df = pd.DataFrame(sessionList[1])
flow_pca2_norm = (flow_df - flow_df.mean()) / (flow_df.max() - flow_df.min())
time_pca_norm = (time_df - time_df.mean()) / (time_df.max() - time_df.min())
features_list = []
features_list.append(list(flow_pca2_norm[0].values))
features_list.append(list(time_pca_norm[0].values))
features_list = np.transpose(features_list)

## apply agglomerative clustering using euclidean distance
X = features_list
# generate the linkage matrix# gener 
Z = linkage(X, 'average')  # method 'average' is chosen based on the cophenetic correlation coefficient, metric is euclidean by default

# calculate the cophenetic correlation coefficient of the clusteing.
# it compares the actual pairwise distances of all samples to those implied by the clustering
# the closer the values is to 1, the better the clustering preserves the original distances
c, coph_dists = cophenet(Z, pdist(X))  # the distance metric in pdist() is euclidean by default
print(c)

## plot dendrogram
def fancy_dendrogram(*args, **kwargs):
    max_d = kwargs.pop('max_d', None)
    if max_d and 'color_threshold' not in kwargs:
        kwargs['color_threshold'] = max_d
    annotate_above = kwargs.pop('annotate_above', 0)

    ddata = dendrogram(*args, **kwargs)

    if not kwargs.get('no_plot', False):
        plt.title('Hierarchical Clustering Dendrogram (truncated)')
        plt.xlabel('sample index or (cluster size)')
        plt.ylabel('distance')
        for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > annotate_above:
                plt.plot(x, y, 'o', c=c)
                plt.annotate("%.3g" % y, (x, y), xytext=(0, -5),
                             textcoords='offset points',
                             va='top', ha='center')
        if max_d:
            plt.axhline(y=max_d, c='k')
    return ddata

plt.figure(figsize=(50, 10))
plt.title('Hierarchical Clustering Dendrogram')
plt.xlabel('users')
plt.ylabel('distance')
fancy_dendrogram(
    Z,
    leaf_rotation=90.,  # rotates the x axis labels
    leaf_font_size=8.,  # font size for the x axis labels
    #annotate_above=10,  # only show annotate bigger than 10.useful in small plots so annotations don't overlap
)
plt.show()

## use elbow method to calculate the number of clusters
last = Z[-8:, 2]  # the distances of last 10 merges
last_rev = last[::-1]  # sort the distances by values, the first one is the biggest one
idxs = np.arange(1, len(last) + 1,step=1)

fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(idxs, last_rev,'b-')  # the blue line, which is the distances of last 10 merges

acceleration = np.diff(last, 2)  # 2nd derivative of the distances， which is the speed at which the tangent slope changes
acceleration_rev = acceleration[::-1]
ax2.plot(idxs[:-2] + 1, acceleration_rev,'g-')  # the green line, which is the acceleration of distance growth

ax1.set_xlabel('clusters')
ax1.set_ylabel('distances of last 8 merges', color='b')
ax2.set_ylabel('2nd derivative of distances', color='g')

plt.xticks(idxs)
plt.title("Sessions data Elbow method result")
plt.grid()
plt.show()

# the number of the clusters is the elbow of blue line and the highest point of green line
k = acceleration_rev.argmax() + 2  # if idx 0 is the max of this we want 2 clusters
print("clusters:", k)

## retrieve clusters
# retrieve clusters by max distance
#clusters = fcluster(Z, max_d, criterion='distance')

# retrieve clusters by the number of clusters
k = 6
clusters = fcluster(Z, k, criterion='maxclust')

# name cluster id from 0
for i in range(len(clusters)):
    clusters[i] = clusters[i] - 1
print(clusters)

## plot clustering result
colors = []
for i in range(len(clusters)):
    if(clusters[i]==0):
        color = 'purple'
    elif(clusters[i]==1):
        color = 'deepskyblue'
    elif(clusters[i]==2):
        color = 'green'
    elif(clusters[i]==3):
        color = 'red'
    elif(clusters[i]==4):
        color = 'magenta'
    elif(clusters[i]==5):
        color = 'blue'
    elif(clusters[i]==6):
        color = 'grey'
    elif(clusters[i]==7):
        color = 'gold'
    colors.append(color)
    
plt.title("Sessions data Agglomerative clustering result ("+str(datetime.now().date())+")")
plt.xlabel('userflow')
plt.ylabel('timespent')
patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3,patch4,patch5],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend
plt.scatter(features_list[:, 0], features_list[:, 1], c=colors)

## get total visits of each user
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "totalVisits" + ".txt"
user_totalVisits = []
with open(path+filename,"r") as file:
    totalVisits = eval(file.readline())
    user_totalVisits.extend(totalVisits)
print(user_totalVisits)

## add clusters to features_list
df_c = pd.DataFrame(features_list)
df_c['cluster'] = clusters

## get clustering results of each user's sessions
cluster_user = []
starting_index = 0
print(user_totalVisits)
for i in range(len(user_totalVisits)):
    totalVisits = user_totalVisits[i]
    df = df_c.iloc[starting_index:starting_index+totalVisits]
    cluster = list(df['cluster'].values)
    cluster_user.append(cluster)
    starting_index = starting_index + totalVisits
print(len(cluster_user))
print(cluster_user)

## plot user-session-cluster
#users = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
users = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

plt.figure(figsize=(15,10))
plt.title("Users' sessions clustering result ("+str(datetime.now().date())+")")
plt.xlabel('user')
plt.ylabel('session')
plt.yticks(np.arange(0,10,step=1))

patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3,patch4,patch5],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend

for i in range(len(cluster_user)):
    for j in range(len(cluster_user[i])):
        #plt.scatter(users[i],cluster_user[i][j])
        if(cluster_user[i][j]==0):
            plt.scatter(users[i],j,c='purple')
        elif(cluster_user[i][j]==1):
            plt.scatter(users[i],j,c='deepskyblue')
        elif(cluster_user[i][j]==2):
            plt.scatter(users[i],j,c='green')
        elif(cluster_user[i][j]==3):
            plt.scatter(users[i],j,c='red')
        elif(cluster_user[i][j]==4):
            plt.scatter(users[i],j,c='magenta')
        elif(cluster_user[i][j]==5):
            plt.scatter(users[i],j,c='blue')
            
## show each user's cluster distribution
df_u = pd.DataFrame(cluster_user)

# show the representative cluster of each user
mode_user = []
for i in range(len(cluster_user)):
    data = Counter(cluster_user[i])
    mode = data.most_common(1)
    mode_user.append(mode)
df_u['mode'] = pd.DataFrame(mode_user)

df_u['user'] = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
#df_u['user'] = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

df_u
