import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import cophenet
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import fcluster
import matplotlib.patches as mpatches

## concatenate all matrices
path = "C:/Users/heyu/PycharmProjects/backup/big matrix records/"
oldBigMatrix = []
for i in range(4):  # 4 files
    filename = "bigMatrix_studyTestUsers_" + str(i) + ".txt"
    with open(path+filename,"r") as file:
        matrix = eval(file.readline())
        oldBigMatrix.extend(matrix)
print(np.array(oldBigMatrix).shape)

# remove duplicated sessions
bigMatrix = []
for i in oldBigMatrix:
    if(i not in bigMatrix):
        bigMatrix.append(i)
print(np.array(bigMatrix).shape)

visitors = len(bigMatrix)
sessions = len(bigMatrix[0])
print(visitors,sessions)

## use pca to reduce dimensions
pca_3d = []
for i in range(visitors):
    pca_list = []
    for j in range(sessions):
        userflow = bigMatrix[i][j]
        pca = PCA(n_components=1)
        principalComponents = pca.fit_transform(userflow)
        df = pd.DataFrame(data = principalComponents)
        df = df[0].values.tolist()
        
        pca_list.append(df)
    pca_3d.append(pca_list)  # till here every 33-list in pca_3d is corresponding to the userflow matrix

# apply pca again
pca1_list =[]
for m in range(visitors):
    pca1 = PCA(n_components=1)
    principalComponents1 = pca1.fit_transform(pca_3d[m])
    df1 = pd.DataFrame(data = principalComponents1)
    df1 = df1[0].values.tolist()
    pca1_list.append(df1)

# round numbers to 2 decimals. 
# the cluster results of 5 decimals, 3 decimals and before rounding are the same
# the heatmap is created by around 2 decimals
for p in range(len(pca1_list)):
    for q in range(len(pca1_list[p])):
        pca1_list[p][q] = round(pca1_list[p][q],2)
#print(pca1_list)
pca1_list = np.array(pca1_list)

## heatmap of all users
x_labels = [str('s'+ str(i)) for i in range(sessions)]
y_labels = [str('u'+ str(i)) for i in range(visitors)]
df_cm = pd.DataFrame(pca1_list,y_labels, x_labels)
plt.figure(figsize=(10,35))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-5,vmax=25)# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()

## apply agglomerative clustering using euclidean distance
X = pca1_list
# generate the linkage matrix# gener 
Z = linkage(X, 'average')  # method 'average' is chosen based on the cophenetic correlation coefficient, metric is euclidean by default

# calculate the cophenetic correlation coefficient of the clusteing.
# it compares the actual pairwise distances of all samples to those implied by the clustering
# the closer the values is to 1, the better the clustering preserves the original distances
c, coph_dists = cophenet(Z, pdist(X))  # the distance metric in pdist() is euclidean by default
print(c)

## plot dendrogram
def fancy_dendrogram(*args, **kwargs):
    max_d = kwargs.pop('max_d', None)
    if max_d and 'color_threshold' not in kwargs:
        kwargs['color_threshold'] = max_d
    annotate_above = kwargs.pop('annotate_above', 0)

    ddata = dendrogram(*args, **kwargs)

    if not kwargs.get('no_plot', False):
        plt.title('Hierarchical Clustering Dendrogram (truncated)')
        plt.xlabel('sample index or (cluster size)')
        plt.ylabel('distance')
        for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > annotate_above:
                plt.plot(x, y, 'o', c=c)
                plt.annotate("%.3g" % y, (x, y), xytext=(0, -5),
                             textcoords='offset points',
                             va='top', ha='center')
        if max_d:
            plt.axhline(y=max_d, c='k')
    return ddata

plt.figure(figsize=(25, 10))
plt.title('Hierarchical Clustering Dendrogram')
plt.xlabel('users')
plt.ylabel('distance')
fancy_dendrogram(
    Z,
    labels=[str('u'+ str(i)) for i in range(visitors)],
    leaf_rotation=90.,  # rotates the x axis labels
    leaf_font_size=12.,  # font size for the x axis labels
    #annotate_above=10,  # only show annotate bigger than 10.useful in small plots so annotations don't overlap
)
plt.show()

## use elbow method to calculate the number of clusters
last = Z[-16:, 2]  # the distances of last 16 merges
last_rev = last[::-1]  # sort the distances by values, the first one is the biggest one
idxs = np.arange(1, len(last) + 1,step=1)

fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(idxs, last_rev,'b-')  # the blue line, which is the distances of last 10 merges

acceleration = np.diff(last, 2)  # 2nd derivative of the distances， which is the speed at which the tangent slope changes
acceleration_rev = acceleration[::-1]
ax2.plot(idxs[:-2] + 1, acceleration_rev,'g-')  # the green line, which is the acceleration of distance growth

ax1.set_xlabel('clusters')
ax1.set_ylabel('distances of last 16 merges', color='b')
ax2.set_ylabel('2nd derivative of distances', color='g')

plt.xticks(idxs)
plt.title("User flow data Elbow method result")
plt.show()

# the number of the clusters is the elbow of blue line and the highest point of green line
k = acceleration_rev.argmax() + 2  # if idx 0 is the max of this we want 2 clusters
print("clusters:", k)

## retrieve clusters
# retrieve clusters by max distance
#max_d = 11
#clusters = fcluster(Z, max_d, criterion='distance')

# retrieve clusters by the number of clusters
k = 7
clusters = fcluster(Z, k, criterion='maxclust')
print(clusters)

## heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == 7]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 6 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-4,vmax=16)# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()
