import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import cophenet
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import fcluster
import time
from time import mktime
from datetime import timedelta
import matplotlib.patches as mpatches
import plotly
plotly.offline.init_notebook_mode()

##choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','amar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get csv data and put it into a 3D list -- features_list, which is used in the clustering method
path = 'C:/Users/heyu/PycharmProjects/content data'   
activity_list = [[0]*3 for a in range(usersNum)]
features_list = [[0]*3 for a in range(usersNum)]

# get three types of activity data to construct features
for i in range(usersNum):    
    filename = '/' + usersName[i] + '_activity_data.csv'
    df = pd.read_csv(path+filename)
    steps = df['completed'].values
    goal = df['suggested'].values
    timestamp = df['datecompleted'].values
    activity_list[i][0] = steps
    activity_list[i][1] = goal
    activity_list[i][2] = timestamp

# calculate the average time difference in days between two activities -- feature 1
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    timestamp = activity_list[i][2]
    for j in range(len(timestamp)):
        ts = int(str(timestamp[j])[:-3])  # remove the last three digits from timestamp to make it specific to seconds
        timestamp[j] = ts
    
    # convert timestamp into dates
    dates = []
    for j in range(len(timestamp)):
        construct_time =time.gmtime(timestamp[j])  # turn timestamp into gmtime
        date = datetime.fromtimestamp(mktime(construct_time)).date()  # turn gmtime into datetime and remove time
        dates.append(date)
    
    if(steps.any()==0):  # 0 activity
        avg = 365  # the bigger avg is, the less frequent the activity is 
    else:
        # remove the date with 0 steps
        new_dates = []
        for j in range(len(steps)):
            if(steps[j]!=0):
                new_dates.append(dates[j])
        new_dates.sort()  # sort the dates

        if(len(new_dates)==1):  # only 1 activity
            avg = 100
        else:
            # calculate the time difference in days between two activities
            time_difference = []
            for j in range(len(new_dates)-1):
                difference = new_dates[j+1] - new_dates[j]
                difference = difference / timedelta(days=1)  # turn timedelta into float
                time_difference.append(difference)

            # calculate the average time difference
            sum_d = 0
            for j in range(len(time_difference)):
                sum_d = sum_d + time_difference[j]
            avg = round(sum_d / len(time_difference),2)
    features_list[i][0] = avg

# calculate the complete percentage -- feature 2
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    goal = activity_list[i][1]
    complete_num = 0
    for j in range(len(steps)):    
        if(steps[j]>=goal[j]):
            complete_num += 1
    complete_ratio = round(complete_num/len(goal),2)
    features_list[i][1] = complete_ratio

# calculate the average steps -- feature 3
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    steps_sum = 0
    if(steps.any()==0):
        steps_avg = 0
    else:
        for j in range(len(steps)):
            steps_sum += steps[j]
        steps_avg = round(steps_sum / len(steps),2)
    features_list[i][2] = steps_avg

features_list = np.array(features_list)

## apply agglomerative clustering using euclidean distance
X = features_list
# generate the linkage matrix# gener 
Z = linkage(X, 'average')  # method 'average' is chosen based on the cophenetic correlation coefficient, metric is euclidean by default

# calculate the cophenetic correlation coefficient of the clusteing.
# it compares the actual pairwise distances of all samples to those implied by the clustering
# the closer the values is to 1, the better the clustering preserves the original distances
c, coph_dists = cophenet(Z, pdist(X))  # the distance metric in pdist() is euclidean by default
print(c)

## plot dendrogram
def fancy_dendrogram(*args, **kwargs):
    max_d = kwargs.pop('max_d', None)
    if max_d and 'color_threshold' not in kwargs:
        kwargs['color_threshold'] = max_d
    annotate_above = kwargs.pop('annotate_above', 0)

    ddata = dendrogram(*args, **kwargs)

    if not kwargs.get('no_plot', False):
        plt.title('Hierarchical Clustering Dendrogram (truncated)')
        plt.xlabel('sample index or (cluster size)')
        plt.ylabel('distance')
        for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > annotate_above:
                plt.plot(x, y, 'o', c=c)
                plt.annotate("%.3g" % y, (x, y), xytext=(0, -5),
                             textcoords='offset points',
                             va='top', ha='center')
        if max_d:
            plt.axhline(y=max_d, c='k')
    return ddata

plt.figure(figsize=(25, 10))
plt.title('Hierarchical Clustering Dendrogram')
plt.xlabel('users')
plt.ylabel('distance')
fancy_dendrogram(
    Z,
    labels=['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10'],
    leaf_rotation=90.,  # rotates the x axis labels
    leaf_font_size=12.,  # font size for the x axis labels
    #annotate_above=10,  # only show annotate bigger than 10.useful in small plots so annotations don't overlap
)
plt.show()

## use elbow method to calculate the number of clusters
last = Z[-10:, 2]  # the distances of last 10 merges
last_rev = last[::-1]  # sort the distances by values, the first one is the biggest one
idxs = np.arange(1, len(last) + 1,step=1)

fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(idxs, last_rev,'b-')  # the blue line, which is the distances of last 10 merges

acceleration = np.diff(last, 2)  # 2nd derivative of the distances， which is the speed at which the tangent slope changes
acceleration_rev = acceleration[::-1]
ax2.plot(idxs[:-2] + 1, acceleration_rev,'g-')  # the green line, which is the acceleration of distance growth

ax1.set_xlabel('clusters')
ax1.set_ylabel('distances of last 10 merges', color='b')
ax2.set_ylabel('2nd derivative of distances', color='g')

plt.xticks(idxs)
plt.title("Activity data Elbow method result")
plt.show()

# the number of the clusters is the elbow of blue line and the highest point of green line
k = acceleration_rev.argmax() + 2  # if idx 0 is the max of this we want 2 clusters
print("clusters:", k)

## retrieve clusters
# retrieve clusters by max distance
#clusters = fcluster(Z, max_d, criterion='distance')

# retrieve clusters by the number of clusters
k = 2
clusters = fcluster(Z, k, criterion='maxclust')
print(clusters)

## plot clustering results with shape 
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = clusters

cluster0 = df_cm.loc[df_cm['cluster'] == 1]
cluster1 = df_cm.loc[df_cm['cluster'] == 2]
cluster2 = df_cm.loc[df_cm['cluster'] == 3]
cluster3 = df_cm.loc[df_cm['cluster'] == 4]
cluster4 = df_cm.loc[df_cm['cluster'] == 5]
cluster5 = df_cm.loc[df_cm['cluster'] == 6]

scatter0 = dict(
    mode = "markers",
    name = "Cluster 0",
    type = "scatter3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    marker = dict( size=2, color='orange')
)
scatter1 = dict(
    mode = "markers",
    name = "Cluster 1",
    type = "scatter3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    marker = dict( size=2, color='blue')
)
scatter2 = dict(
    mode = "markers",
    name = "Cluster 2",
    type = "scatter3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    marker = dict( size=2, color='red')
)
scatter3 = dict(
    mode = "markers",
    name = "Cluster 3",
    type = "scatter3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    marker = dict( size=2, color='green')
)
scatter4 = dict(
    mode = "markers",
    name = "Cluster 4",
    type = "scatter3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    marker = dict( size=2, color='grey')
)
scatter5 = dict(
    mode = "markers",
    name = "Cluster 5",
    type = "scatter3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    marker = dict( size=2, color='magenta')
)
cluster0 = dict(
    alphahull = 5,
    name = "Cluster 0",
    opacity = .1,
    type = "mesh3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    color='orange', showscale = True
)
cluster1 = dict(
    alphahull = 5,
    name = "Cluster 1",
    opacity = .1,
    type = "mesh3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    color='blue', showscale = True
)
cluster2 = dict(
    alphahull = 5,
    name = "Cluster 2",
    opacity = .1,
    type = "mesh3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    color='red', showscale = True
)
cluster3 = dict(
    alphahull = 5,
    name = "Cluster 3",
    opacity = .1,
    type = "mesh3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    color='green', showscale = True
)
cluster4 = dict(
    alphahull = 5,
    name = "Cluster 4",
    opacity = .1,
    type = "mesh3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    color='grey', showscale = True
)
cluster5 = dict(
    alphahull = 5,
    name = "Cluster 5",
    opacity = .1,
    type = "mesh3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    color='magenta', showscale = True
)
layout = dict(
    title = 'Activity data Agglomerative clustering result ('+str(datetime.now().date())+")",
    scene = dict(
        xaxis = dict( title='avg_days',zeroline=True ),
        yaxis = dict( title='complete_ratio',zeroline=True ),
        zaxis = dict( title='avg_steps',zeroline=True ),
    )
)
fig = dict( data=[scatter0, scatter1, scatter2, scatter3, scatter4,scatter5, cluster0, cluster1, cluster2, cluster3, cluster4,cluster5], layout=layout )
plotly.offline.iplot(fig)

## normalize features
features_list_transpose = np.array(features_list).transpose()
avg_days_list = features_list_transpose[0]
complete_ratio_list = features_list_transpose[1]
avg_steps_list = features_list_transpose[2]

avg_days_df = pd.DataFrame(avg_days_list)
complete_ratio_df = pd.DataFrame(complete_ratio_list)
avg_steps_df = pd.DataFrame(avg_steps_list)
avg_days_norm = (avg_days_df - avg_days_df.mean()) / (avg_days_df.max() - avg_days_df.min())
complete_ratio_norm = (complete_ratio_df - complete_ratio_df.mean()) / (complete_ratio_df.max() - complete_ratio_df.min())
avg_steps_norm = (avg_steps_df - avg_steps_df.mean()) / (avg_steps_df.max() - avg_steps_df.min())
features_list_transpose[0] = list(avg_days_norm[0].values)
features_list_transpose[1] = list(complete_ratio_norm[0].values)
features_list_transpose[2] = list(avg_steps_norm[0].values)

features_list = features_list_transpose.transpose()

# round to 2 decimals
for p in range(len(features_list)):
    for q in range(len(features_list[p])):
        features_list[p][q] = round(features_list[p][q],2)
        
## plot heatmap of all users
x_labels = ['avg_days','complete_ratio','avg_steps']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-1,vmax=1)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == 1]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 0 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-1,vmax=1)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
