import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import cophenet
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import fcluster
import matplotlib.patches as mpatches

## choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','amar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get csv data and put it into a matrix and do feature extraction. columns are features. rows are users
path = 'C:/Users/heyu/PycharmProjects/content data'
baseline_matrix = [['0']*5 for a in range(usersNum)]
# get five features from csv files
for i in range(usersNum):    
    filename = '/' + usersName[i] + '_baseline_data.csv'
    df = pd.read_csv(path+filename)
    age = df['answer'][0]
    gender = df['answer'][1]
    employment = df['answer'][2]
    pain_current = df['answer'][3]
    rmdq = df['answer'][4]
    baseline_matrix[i][0] = age
    baseline_matrix[i][1] = gender
    baseline_matrix[i][2] = employment
    baseline_matrix[i][3] = pain_current
    baseline_matrix[i][4] = rmdq

# feature extraction for age
for i in range(len(baseline_matrix)):
    age = int(baseline_matrix[i][0])
    if(age<25):
        baseline_matrix[i][0] = 0
    elif(25<=age<55):
        baseline_matrix[i][0] = 1
    elif(55<=age<65):
        baseline_matrix[i][0] = 2
    elif(65<=age):
        baseline_matrix[i][0] = 3
    else:
        baseline_matrix[i][0] = -1

# feature extraction for gender
for i in range(len(baseline_matrix)):
    gender = baseline_matrix[i][1]
    if(gender=='male'):
        baseline_matrix[i][1] = 0
    elif(gender=='female'):
        baseline_matrix[i][1] = 1
    else:
        baseline_matrix[i][1] = -1

# feature extraction for employment
for i in range(len(baseline_matrix)):
    employment = baseline_matrix[i][2]
    if(employment=='Full-time'):
        baseline_matrix[i][2] = 0
    elif(employment=='Part-time'):
        baseline_matrix[i][2] = 1
    elif(employment=='unemployed'):
        baseline_matrix[i][2] = 2
    elif(employment=='other'):
        baseline_matrix[i][2] = 3
    elif(employment=='full-time-housework'):
        baseline_matrix[i][2] = 4
    elif(employment=='retired'):
        baseline_matrix[i][2] = 5
    elif(employment=='military'):
        baseline_matrix[i][2] = 6
    else:
        baseline_matrix[i][2] = -1

# feature extraction for pain_current
for i in range(len(baseline_matrix)):
    pain_current = baseline_matrix[i][3]
    if(pain_current=='1week'):
        baseline_matrix[i][3] = 0
    elif(pain_current=='4weeks'):
        baseline_matrix[i][3] = 1
    elif(pain_current=='12weeks'):
        baseline_matrix[i][3] = 2
    elif(pain_current=='Above12weeks'):
        baseline_matrix[i][3] = 3
    else:
        baseline_matrix[i][3] = -1

# feature extraction for rmdq
for i in range(len(baseline_matrix)):
    rmdq = int(baseline_matrix[i][4])
    if(rmdq<5):
        baseline_matrix[i][4] = 0
    elif(5<=rmdq<10):
        baseline_matrix[i][4] = 1
    elif(10<=rmdq<15):
        baseline_matrix[i][4] = 2
    elif(15<=rmdq<20):
        baseline_matrix[i][4] = 3
    elif(20<=rmdq<25):
        baseline_matrix[i][4] = 4
    else:
        baseline_matrix[i][4] = -1

features_list = np.array(baseline_matrix)

## heatmap of all users
x_labels = ['age','gender','employment','pain_current','rmdq']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(baseline_matrix, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=0,vmax=4)# font size
plt.show()

## apply agglomerative clustering using euclidean distance
X = baseline_matrix
# generate the linkage matrix# gener 
Z = linkage(X, 'average')  # method 'average' is chosen based on the cophenetic correlation coefficient, metric is euclidean by default

# calculate the cophenetic correlation coefficient of the clusteing.
# it compares the actual pairwise distances of all samples to those implied by the clustering
# the closer the values is to 1, the better the clustering preserves the original distances
c, coph_dists = cophenet(Z, pdist(X))  # the distance metric in pdist() is euclidean by default
print(c)

## plot dendrogram
def fancy_dendrogram(*args, **kwargs):
    max_d = kwargs.pop('max_d', None)
    if max_d and 'color_threshold' not in kwargs:
        kwargs['color_threshold'] = max_d
    annotate_above = kwargs.pop('annotate_above', 0)

    ddata = dendrogram(*args, **kwargs)

    if not kwargs.get('no_plot', False):
        for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > annotate_above:
                plt.plot(x, y, 'o', c=c)
                plt.annotate("%.3g" % y, (x, y), xytext=(0, -5),
                             textcoords='offset points',
                             va='top', ha='center')
        if max_d:
            plt.axhline(y=max_d, c='k')
    return ddata

plt.figure(figsize=(10, 6))
plt.title('Baseline data Agglomerative Clustering Dendrogram')
plt.xlabel('users')
plt.ylabel('distance')
fancy_dendrogram(
    Z,
    labels = [str('u'+ str(i)) for i in range(usersNum)],
    #labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10'],
    leaf_rotation=90.,  # rotates the x axis labels
    leaf_font_size=12.,  # font size for the x axis labels
    #annotate_above=10,  # only show annotate bigger than 10.useful in small plots so annotations don't overlap
)
plt.show()

## use elbow method to calculate the number of clusters
last = Z[-10:, 2]  # the distances of last 10 merges
last_rev = last[::-1]  # sort the distances by values, the first one is the biggest one
idxs = np.arange(1, len(last) + 1,step=1)

fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(idxs, last_rev,'b-')  # the blue line, which is the distances of last 10 merges

acceleration = np.diff(last, 2)  # 2nd derivative of the distances， which is the speed at which the tangent slope changes
acceleration_rev = acceleration[::-1]
ax2.plot(idxs[:-2] + 1, acceleration_rev,'g-')  # the green line, which is the acceleration of distance growth

ax1.set_xlabel('clusters')
ax1.set_ylabel('distances of last 10 merges', color='b')
ax2.set_ylabel('2nd derivative of distances', color='g')

plt.xticks(idxs)
plt.title("Baseline data Elbow method result")
plt.show()

# the number of the clusters is the elbow of blue line and the highest point of green line
k = acceleration_rev.argmax() + 2  # if idx 0 is the max of this we want 2 clusters
print("clusters:", k)

## retrieve clusters
# retrieve clusters by max distance
#clusters = fcluster(Z, max_d, criterion='distance')

# retrieve clusters by the number of clusters
k = 2
clusters = fcluster(Z, k, criterion='maxclust')
print(clusters)

## heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == 2]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 1 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=0,vmax=4)# font size
plt.show()
