import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import cophenet
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import fcluster
import time
from time import mktime
from datetime import timedelta
from itertools import groupby
import matplotlib.patches as mpatches
import plotly
plotly.offline.init_notebook_mode()

## choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','annelovise_no']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get exercise list
exerciseId_list = ['back_01_01', 'back_01_02', 'back_02_01', 'back_02_02', 'back_03_01', 'back_03_02', 'back_04_01', 'back_04_02', 'back_05_01', 'back_05_02', 'back_06_01', 'back_06_02', 'glut_01_01', 'glut_01_02', 'glut_02_01', 'glut_02_02', 'glut_03_01', 'glut_03_02', 'glut_04_01', 'glut_04_02', 'glut_05_01', 'glut_05_02', 'glut_06_01', 'glut_06_02', 'ab_01_01', 'ab_01_02', 'ab_02_01', 'ab_02_02', 'ab_03_01', 'ab_03_02', 'ab_03_03', 'ab_04_01', 'ab_04_02', 'ab_04_03', 'ab_05_01', 'ab_05_02', 'ab_06_01', 'ab_06_02', 'core_01_01', 'core_01_02', 'core_02_01', 'core_02_02', 'core_03_01', 'core_03_02', 'core_04_01', 'core_04_02', 'core_05_01', 'core_05_02', 'core_06_01', 'core_06_02', 'pain_01_01', 'pain_01_02', 'pain_02_01', 'pain_02_02', 'pain_02_03', 'pain_02_04', 'pain_03_01', 'pain_03_02', 'pain_03_03', 'pain_04_01', 'pain_04_02', 'flex_01_01', 'flex_01_02', 'flex_01_03', 'flex_01_04', 'flex_01_05', 'flex_01_06', 'flex_01_07', 'flex_01_08', 'flex_01_09']
print(len(exerciseId_list))

## get csv data and put it into a matrix with 3 features
path = 'C:/Users/heyu/PycharmProjects/content data'   
features_list = []
exercise_freq_matrix = []
avg_num_exercise_list = []
complete_ratio_list = []
for i in range(usersNum):
    # get completed data
    filename1 = '/' + usersName[i] + '_exercise_data.csv'
    df1 = pd.read_csv(path+filename1)
    exerciseIdCom = df1['exerciseid'].values
    timestampCom = df1['datecompleted'].values
    timestampCom.sort()
    tsCom_uni = []
    for j in range(len(timestampCom)):
        if(not(timestampCom[j] in tsCom_uni)):
            tsCom_uni.append(timestampCom[j])
    
    # get suggested data
    filename2 = '/' + usersName[i] + '_exercise_data_suggested.csv'
    df2 = pd.read_csv(path+filename2)
    exerciseIdSug = df2['exerciseid'].values
    timestampSug = df2['datesuggested'].values
    timestampSug.sort()
    tsSug_uni = []
    for j in(range(len(timestampSug))):
        if(not(timestampSug[j] in tsSug_uni)):
            tsSug_uni.append(timestampSug[j])
    
    # remove duplicated timestamp
    duplicate = 0
    for j in range(len(tsSug_uni)-1):
        if((tsSug_uni[j+1]-tsSug_uni[j])<1000):
            duplicate = tsSug_uni[j]
    df2 = df2[df2.datesuggested != duplicate]
    timestampSug = df2['datesuggested'].values
    timestampSug.sort()
    tsSug_uni = []
    for j in(range(len(timestampSug))):
        if(not(timestampSug[j] in tsSug_uni)):
            tsSug_uni.append(timestampSug[j])
    
    # use exercise ids in all sessions as feature 1, showing frequency information
    exercise_freq_list = [0]*len(exerciseId_list)
    for j in exerciseIdCom:
        index_j = exerciseId_list.index(j)
        exercise_freq_list[index_j] +=1
    exercise_freq_matrix.append(exercise_freq_list)

    # separate sessions
    sessionsCom = [list(g) for k, g in groupby(timestampCom)]
    sessionsSug = [list(g) for k, g in groupby(timestampSug)]
    
    # number of exercises in one session
    sessionCom_num = []
    for j in range(len(sessionsCom)):
        sessionCom_num.append(len(sessionsCom[j]))
    sessionSug_num = []
    for j in range(len(sessionsSug)):
        sessionSug_num.append(len(sessionsSug[j]))
    
    # calculate average number of exercises in one session (weeks with 0 exercises are not included) -- feature 2
    sum_num_exercise = 0
    for j in range(len(sessionCom_num)):
        sum_num_exercise = sum_num_exercise + sessionCom_num[j]
    avg_num_exercise = sum_num_exercise / len(sessionCom_num)
    avg_num_exercise_list.append(avg_num_exercise)
    
    # complete ratio of each week -- feature 3
    # calculate the expected timestamp
    tsExp = []
    for j in range(len(tsSug_uni)):
        ts_exp = tsSug_uni[j] + 604800000  # add 7 days
        tsExp.append(ts_exp)
    
    # get the suggested timestamp with exercises completed in that week
    tsSugCom = []
    for j in range(len(tsCom_uni)):
        for k in range(len(tsExp)):
            if(abs(tsCom_uni[j]-tsExp[k])<86400000):  # smaller than 1 day
                tsSugCom.append([tsCom_uni[j],tsSug_uni[k]])  # put the completed timestamp and the corresponding suggested timestamp into a pair
    
    # complete_ratio 
    complete_ratio = [0]*len(sessionSug_num)
    if(len(sessionSug_num)==len(sessionCom_num)):  # if every week there are exercises completed
        for j in range(len(sessionSug_num)):
            complete_ratio[j] = sessionCom_num[j] / sessionSug_num[j]  # for weeks with exercises cpmpleted
    else:  # if there is a week or weeks with no exercise completed
        for j in tsSugCom:
            index_sug = tsSug_uni.index(j[1])
            index_com = tsCom_uni.index(j[0])
            complete_ratio[index_sug] = sessionCom_num[index_com] / sessionSug_num[index_sug]  # put non_zero complete ratio to the weeks with exercises completed
    complete_ratio_list.append(complete_ratio)

# apply pca to exercise_freq_matrix
pca = PCA(n_components=1)
principalComponents0 = pca.fit_transform(exercise_freq_matrix)
exercise_freq_users = pd.DataFrame(data = principalComponents0)
exercise_freq_users = exercise_freq_users[0].values.tolist()

# pad 0s to complete ratio list to make matrix
pad = len(max(complete_ratio_list,key=len))
complete_ratio_list = np.array([i + [0]*(pad-len(i)) for i in complete_ratio_list])
# apply pca on complete_ratio_matrix
pca = PCA(n_components=1)
principalComponents1 = pca.fit_transform(complete_ratio_list)
complete_ratio_users = pd.DataFrame(data = principalComponents1)
complete_ratio_users = complete_ratio_users[0].values.tolist()

features_list.append(exercise_freq_users)
features_list.append(avg_num_exercise_list)
features_list.append(complete_ratio_users)

# transpose the features list
features_list = np.transpose(features_list)

## apply agglomerative clustering using euclidean distance
X = features_list
# generate the linkage matrix# gener 
Z = linkage(X, 'average')  # method 'average' is chosen based on the cophenetic correlation coefficient, metric is euclidean by default

# calculate the cophenetic correlation coefficient of the clusteing.
# it compares the actual pairwise distances of all samples to those implied by the clustering
# the closer the values is to 1, the better the clustering preserves the original distances
c, coph_dists = cophenet(Z, pdist(X))  # the distance metric in pdist() is euclidean by default
print(c)

## plot dendrogram
def fancy_dendrogram(*args, **kwargs):
    max_d = kwargs.pop('max_d', None)
    if max_d and 'color_threshold' not in kwargs:
        kwargs['color_threshold'] = max_d
    annotate_above = kwargs.pop('annotate_above', 0)

    ddata = dendrogram(*args, **kwargs)

    if not kwargs.get('no_plot', False):
        for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > annotate_above:
                plt.plot(x, y, 'o', c=c)
                plt.annotate("%.3g" % y, (x, y), xytext=(0, -5),
                             textcoords='offset points',
                             va='top', ha='center')
        if max_d:
            plt.axhline(y=max_d, c='k')
    return ddata

plt.figure(figsize=(25, 10))
plt.title('Hierarchical Clustering Dendrogram')
plt.xlabel('users')
plt.ylabel('distance')
fancy_dendrogram(
    Z,
    labels = ['agnar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10'],
    leaf_rotation=90.,  # rotates the x axis labels
    leaf_font_size=12.,  # font size for the x axis labels
    #annotate_above=10,  # only show annotate bigger than 10.useful in small plots so annotations don't overlap
)
plt.show()

## use elbow method to calculate the number of clusters
last = Z[-10:, 2]  # the distances of last 10 merges
last_rev = last[::-1]  # sort the distances by values, the first one is the biggest one
idxs = np.arange(1, len(last) + 1,step=1)

fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(idxs, last_rev,'b-')  # the blue line, which is the distances of last 10 merges

acceleration = np.diff(last, 2)  # 2nd derivative of the distances， which is the speed at which the tangent slope changes
acceleration_rev = acceleration[::-1]
ax2.plot(idxs[:-2] + 1, acceleration_rev,'g-')  # the green line, which is the acceleration of distance growth

ax1.set_xlabel('clusters')
ax1.set_ylabel('distances of last 10 merges', color='b')
ax2.set_ylabel('2nd derivative of distances', color='g')

plt.xticks(idxs)
plt.title("Exercise data Elbow method result")
plt.show()

# the number of the clusters is the elbow of blue line and the highest point of green line
k = acceleration_rev.argmax() + 2  # if idx 0 is the max of this we want 2 clusters
print("clusters:", k)

## retrieve clusters
# retrieve clusters by max distance
#clusters = fcluster(Z, max_d, criterion='distance')

# retrieve clusters by the number of clusters
k = 2
clusters = fcluster(Z, k, criterion='maxclust')
print(clusters)

## plot clustering result with shape
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = clusters

cluster0 = df_cm.loc[df_cm['cluster'] == 1]
cluster1 = df_cm.loc[df_cm['cluster'] == 2]
cluster2 = df_cm.loc[df_cm['cluster'] == 3]
cluster3 = df_cm.loc[df_cm['cluster'] == 4]
cluster4 = df_cm.loc[df_cm['cluster'] == 5]
cluster5 = df_cm.loc[df_cm['cluster'] == 6]

scatter0 = dict(
    mode = "markers",
    name = "Cluster 0",
    type = "scatter3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    marker = dict( size=2, color='green')
)
scatter1 = dict(
    mode = "markers",
    name = "Cluster 1",
    type = "scatter3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    marker = dict( size=2, color='blue')
)
scatter2 = dict(
    mode = "markers",
    name = "Cluster 2",
    type = "scatter3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    marker = dict( size=2, color='red')
)
scatter3 = dict(
    mode = "markers",
    name = "Cluster 3",
    type = "scatter3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    marker = dict( size=2, color='magenta')
)
scatter4 = dict(
    mode = "markers",
    name = "Cluster 4",
    type = "scatter3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    marker = dict( size=2, color='grey')
)
scatter5 = dict(
    mode = "markers",
    name = "Cluster 5",
    type = "scatter3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    marker = dict( size=2, color='yellow')
)
cluster0 = dict(
    alphahull = 5,
    name = "Cluster 0",
    opacity = .1,
    type = "mesh3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    color='green', showscale = True
)
cluster1 = dict(
    alphahull = 5,
    name = "Cluster 1",
    opacity = .1,
    type = "mesh3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    color='blue', showscale = True
)
cluster2 = dict(
    alphahull = 5,
    name = "Cluster 2",
    opacity = .1,
    type = "mesh3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    color='red', showscale = True
)
cluster3 = dict(
    alphahull = 5,
    name = "Cluster 3",
    opacity = .1,
    type = "mesh3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    color='magenta', showscale = True
)
cluster4 = dict(
    alphahull = 5,
    name = "Cluster 4",
    opacity = .1,
    type = "mesh3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    color='grey', showscale = True
)
cluster5 = dict(
    alphahull = 5,
    name = "Cluster 5",
    opacity = .1,
    type = "mesh3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    color='yellow', showscale = True
)
layout = dict(
    title = 'Exercise data Agglomerative clustering result ('+str(datetime.now().date())+")",
    scene = dict(
        xaxis = dict( title='exe_freq',zeroline=True ),
        yaxis = dict( title='avg_num_exe_per_session',zeroline=True ),
        zaxis = dict( title='complete_ratio',zeroline=True ),
    )
)
fig = dict( data=[scatter0, scatter1, scatter2,scatter3,scatter4, scatter5, cluster0, cluster1, cluster2,cluster3,cluster4,cluster5], layout=layout )
plotly.offline.iplot(fig)

## normalize features
exe_freq_df = pd.DataFrame(exercise_freq_users)
avg_num_exe_df = pd.DataFrame(avg_num_exercise_list)
complete_ratio_df = pd.DataFrame(complete_ratio_users)

exe_freq_norm = (exe_freq_df - exe_freq_df.mean()) / (exe_freq_df.max() - exe_freq_df.min())
avg_num_exe_norm = (avg_num_exe_df - avg_num_exe_df.mean()) / (avg_num_exe_df.max() - avg_num_exe_df.min())
complete_ratio_norm = (complete_ratio_df - complete_ratio_df.mean()) / (complete_ratio_df.max() - complete_ratio_df.min())

exercise_freq_users_norm = list(exe_freq_norm[0].values)
avg_num_exercise_list_norm = list(avg_num_exe_norm[0].values)
complete_ratio_users_norm = list(complete_ratio_norm[0].values)

features_list_norm = []
features_list_norm.append(exercise_freq_users_norm)
features_list_norm.append(avg_num_exercise_list_norm)
features_list_norm.append(complete_ratio_users_norm)

# round each number to 2 decimals
for i in range(len(features_list_norm)):
    for j in range(len(features_list_norm[i])):
        features_list_norm[i][j] = round(features_list_norm[i][j],2)
        
# transpose the features list
features_list = np.transpose(features_list_norm)

## heatmap of all users after normalization
x_labels = ['exe_freq','avg_num_exe_per_session','complete_ratio']
#y_labels = ['agnar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-0.4,vmax=0.6)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

# heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == 2]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 0 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-0.6,vmax=0.6)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
