import pandas as pd
import requests
import json
from pandas.io.json import json_normalize
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_samples, silhouette_score
import matplotlib.cm as cm
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
from collections import Counter
import matplotlib.patches as mpatches

## get backup sessions data
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "sessions" + ".txt"
sessionList = []
with open(path+filename,"r") as file:
    session = eval(file.readline())
    sessionList.extend(session)
    
## normalize time spent
flow_df = pd.DataFrame(sessionList[0])
time_df = pd.DataFrame(sessionList[1])
flow_pca2_norm = (flow_df - flow_df.mean()) / (flow_df.max() - flow_df.min())
time_pca_norm = (time_df - time_df.mean()) / (time_df.max() - time_df.min())
features_list = []
features_list.append(list(flow_pca2_norm[0].values))
features_list.append(list(time_pca_norm[0].values))
features_list = np.transpose(features_list)

## silhouette analysis
range_n_clusters = [2, 3, 4, 5, 6, 7, 8]
X = features_list

for n_clusters in range_n_clusters:
    # Create a subplot with 1 row and 2 columns
    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(18, 7)

    # The 1st subplot is the silhouette plot
    # The silhouette coefficient can range from -1, 1 but in this example all
    # lie within [-0.1, 1]
    ax1.set_xlim([-0.1, 1])
    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters, to demarcate them clearly.
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])

    # Initialize the clusterer with n_clusters value and a random generator
    # seed of 10 for reproducibility.
    clusterer = KMeans(n_clusters=n_clusters, random_state=10)
    cluster_labels = clusterer.fit_predict(X)

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    silhouette_avg = silhouette_score(X, cluster_labels)
    print("For n_clusters =", n_clusters,
          "The average silhouette_score is :", silhouette_avg)

    # Compute the silhouette scores for each sample
    sample_silhouette_values = silhouette_samples(X, cluster_labels)

    y_lower = 10
    for i in range(n_clusters):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == i]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        color = cm.spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, ith_cluster_silhouette_values,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples

    ax1.set_title("The silhouette plot for the various clusters.")
    ax1.set_xlabel("The silhouette coefficient values")
    ax1.set_ylabel("Cluster label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

    # 2nd Plot showing the actual clusters formed
    colors = cm.spectral(cluster_labels.astype(float) / n_clusters)
    ax2.scatter(X[:, 0], X[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                c=colors, edgecolor='k')

    # Labeling the clusters
    centers = clusterer.cluster_centers_
    # Draw white circles at cluster centers
    ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                c="white", alpha=1, s=200, edgecolor='k')

    for i, c in enumerate(centers):
        ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                    s=50, edgecolor='k')

    ax2.set_title("The visualization of the clustered data.")
    ax2.set_xlabel("Feature space for the 1st feature")
    ax2.set_ylabel("Feature space for the 2nd feature")

    plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                  "with n_clusters = %d" % n_clusters),
                 fontsize=14, fontweight='bold')

    plt.show()
    
## plot dataset
plt.title("Sessions distribution plot")
plt.xlabel('userflow')
plt.ylabel('timespent')
plt.scatter(features_list[:, 0], features_list[:, 1])

## apply kmeans
# Initializing KMeans
kmeans = KMeans(n_clusters=4)
# Fitting with inputs
# fit computes k-means clustering
kmeans = kmeans.fit(features_list)
# Predicting the clusters
# predict computes cluster centers and predict cluster index for each sample, returns index of the cluster each sample belong to
labels = kmeans.predict(features_list)
# Getting the cluster centers in coordinates
C = kmeans.cluster_centers_
clusters = kmeans.labels_  # for each visitor(data point), get its cluster's id. 
print(clusters)

## plot clustering result
colors = []
for i in range(len(clusters)):
    if(clusters[i]==0):
        color = 'purple'
    elif(clusters[i]==1):
        color = 'deepskyblue'
    elif(clusters[i]==2):
        color = 'green'
    elif(clusters[i]==3):
        color = 'red'
    elif(clusters[i]==4):
        color = 'magenta'
    elif(clusters[i]==5):
        color = 'blue'
    elif(clusters[i]==6):
        color = 'grey'
    elif(clusters[i]==7):
        color = 'gold'
    colors.append(color)

plt.title("Sessions data K-means clustering result")
plt.xlabel('userflow')
plt.ylabel('timespent')

patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend
plt.scatter(features_list[:, 0], features_list[:, 1], c=colors)
#plt.scatter(C[:, 0], C[:, 1], marker='*', c='red', s=100)

## get total visits of each user
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "totalVisits" + ".txt"
user_totalVisits = []
with open(path+filename,"r") as file:
    totalVisits = eval(file.readline())
    user_totalVisits.extend(totalVisits)
print(user_totalVisits)

## add cluster ids
df_c = pd.DataFrame(features_list)
df_c['cluster'] = clusters

## get clustering results of each user's sessions
cluster_user = []
starting_index = 0
print(user_totalVisits)
for i in range(len(user_totalVisits)):
    totalVisits = user_totalVisits[i]
    df = df_c.iloc[starting_index:starting_index+totalVisits]
    cluster = list(df['cluster'].values)
    cluster_user.append(cluster)
    starting_index = starting_index + totalVisits
print(len(cluster_user))
print(cluster_user)

## plot user-session-cluster
#users = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
users = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

plt.figure(figsize=(12,10))
plt.title("Users' sessions clustering result")
plt.xlabel('user')
plt.ylabel('session')
plt.yticks(np.arange(0,10,step=1))

patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend

for i in range(len(cluster_user)):
    for j in range(len(cluster_user[i])):
        #plt.scatter(users[i],cluster_user[i][j])
        if(cluster_user[i][j]==0):
            plt.scatter(users[i],j,c='purple',marker='o')
        elif(cluster_user[i][j]==1):
            plt.scatter(users[i],j,c='deepskyblue',marker='*')
        elif(cluster_user[i][j]==2):
            plt.scatter(users[i],j,c='green',marker='D')
        elif(cluster_user[i][j]==3):
            plt.scatter(users[i],j,c='red',marker='x')
            
## show each user's cluster distribution
df_u = pd.DataFrame(cluster_user)

# show the representative cluster of each user
mode_user = []
for i in range(len(cluster_user)):
    data = Counter(cluster_user[i])
    mode = data.most_common(1)
    mode_user.append(mode)
df_u['mode'] = pd.DataFrame(mode_user)

df_u['user'] = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
#df_u['user'] = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

df_u            
