import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from sklearn.cluster import KMeans
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_samples, silhouette_score
import matplotlib.cm as cm
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime

## concatenate all matrices
path = "C:/Users/heyu/PycharmProjects/backup/big matrix records/"
oldBigMatrix = []
for i in range(4):  # 4 files
    filename = "bigMatrix_studyTestUsers_" + str(i) + ".txt"
    with open(path+filename,"r") as file:
        matrix = eval(file.readline())
        oldBigMatrix.extend(matrix)
print(np.array(oldBigMatrix).shape)

# remove duplicated sessions
bigMatrix = []
for i in oldBigMatrix:
    if(i not in bigMatrix):
        bigMatrix.append(i)
print(np.array(bigMatrix).shape)

visitors = len(bigMatrix)
sessions = len(bigMatrix[0])
print(visitors,sessions)

## use pca to reduce dimensions
pca_3d = []
for i in range(visitors):
    pca_list = []
    for j in range(sessions):
        userflow = bigMatrix[i][j]
        pca = PCA(n_components=1)
        principalComponents = pca.fit_transform(userflow)
        df = pd.DataFrame(data = principalComponents)
        df = df[0].values.tolist()
        
        pca_list.append(df)
    pca_3d.append(pca_list)  # till here every 33-list in pca_3d is corresponding to the userflow matrix

# apply pca again
pca1_list =[]
for m in range(visitors):
    pca1 = PCA(n_components=1)
    principalComponents1 = pca1.fit_transform(pca_3d[m])
    df1 = pd.DataFrame(data = principalComponents1)
    df1 = df1[0].values.tolist()
    pca1_list.append(df1)

# round numbers to 2 decimals. 
# the cluster results of 5 decimals, 3 decimals and before rounding are the same
# the heatmap is created by around 2 decimals
for p in range(len(pca1_list)):
    for q in range(len(pca1_list[p])):
        pca1_list[p][q] = round(pca1_list[p][q],2)
#print(pca1_list)
pca1_list = np.array(pca1_list)

## use pca to reduce 3 features to 2 featrues for silhouette analysis
pca = PCA(n_components=2).fit(pca1_list)
X = pca.transform(pca1_list)  
#print(X)

## calculate the max number of clusters
max_clusters = round(visitors * 0.3,0)
print(max_clusters)

## silhouette analysis

range_n_clusters = [2,3,4,5,6,7,8,9,10,11,12,13,14,15,16]

for n_clusters in range_n_clusters:
    # Create a subplot with 1 row and 2 columns
    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(18, 7)

    # The 1st subplot is the silhouette plot
    # The silhouette coefficient can range from -1, 1 but in this example all
    # lie within [-0.1, 1]
    ax1.set_xlim([-0.1, 1])
    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters, to demarcate them clearly.
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])

    # Initialize the clusterer with n_clusters value and a random generator
    # seed of 10 for reproducibility.
    clusterer = KMeans(n_clusters=n_clusters, random_state=10)
    cluster_labels = clusterer.fit_predict(X)

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    silhouette_avg = silhouette_score(X, cluster_labels)
    print("For n_clusters =", n_clusters,
          "The average silhouette_score is :", silhouette_avg)

    # Compute the silhouette scores for each sample
    sample_silhouette_values = silhouette_samples(X, cluster_labels)

    y_lower = 10
    for i in range(n_clusters):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == i]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        color = cm.spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, ith_cluster_silhouette_values,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples

    ax1.set_title("The silhouette plot for the various clusters.")
    ax1.set_xlabel("The silhouette coefficient values")
    ax1.set_ylabel("Cluster label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

    # 2nd Plot showing the actual clusters formed
    colors = cm.spectral(cluster_labels.astype(float) / n_clusters)
    ax2.scatter(X[:, 0], X[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                c=colors, edgecolor='k')

    # Labeling the clusters
    centers = clusterer.cluster_centers_
    # Draw white circles at cluster centers
    ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                c="white", alpha=1, s=200, edgecolor='k')

    for i, c in enumerate(centers):
        ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                    s=50, edgecolor='k')

    ax2.set_title("The visualization of the clustered data.")
    ax2.set_xlabel("Feature space for the 1st feature")
    ax2.set_ylabel("Feature space for the 2nd feature")

    plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                  "with n_clusters = %d" % n_clusters),
                 fontsize=14, fontweight='bold')

    plt.show()
    
## heatmap of all users
x_labels = [str('s'+ str(i)) for i in range(sessions)]
y_labels = [str('u'+ str(i)) for i in range(visitors)]
df_cm = pd.DataFrame(pca1_list,y_labels, x_labels)
plt.figure(figsize=(10,25))
plt.title("User flow heatmap of all users")
sns.set(font_scale=1.6)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g')# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()

## apply KMeans
pca1_df = pd.DataFrame(pca1_list)

# Initializing KMeans
kmeans = KMeans(n_clusters=4)
# Fitting with inputs
# fit computes k-means clustering
kmeans = kmeans.fit(pca1_list)
# Predicting the clusters
# predict computes cluster centers and predict cluster index for each sample, returns index of the cluster each sample belong to
labels = kmeans.predict(pca1_list)
# Getting the cluster centers in coordinates
C = kmeans.cluster_centers_  # the coordinates of cluster centers
#print(C)
visitor_clusterid = kmeans.labels_  # for each visitor(data point), get its cluster's id. visitor 0 belongs to cluster 0
print(visitor_clusterid)

## heatmap for each cluster
df_cm['cluster'] = visitor_clusterid

df = df_cm.loc[df_cm['cluster'] == 0]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 0 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-4,vmax=16)# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()
