import pandas as pd
import numpy as np
from pandas.plotting import scatter_matrix
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
import seaborn as sns
import time
from time import mktime
from datetime import datetime
from datetime import timedelta
from mpl_toolkits.mplot3d import Axes3D
from sklearn.metrics import silhouette_samples, silhouette_score
import matplotlib.cm as cm
from sklearn.decomposition import PCA
import plotly
plotly.offline.init_notebook_mode()

##choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','amar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get csv data and put it into a 3D list -- features_list, which is used in the clustering method
path = 'C:/Users/heyu/PycharmProjects/content data'   
activity_list = [[0]*3 for a in range(usersNum)]
features_list = [[0]*3 for a in range(usersNum)]

# get three types of activity data to construct features
for i in range(usersNum):    
    filename = '/' + usersName[i] + '_activity_data.csv'
    df = pd.read_csv(path+filename)
    steps = df['completed'].values
    goal = df['suggested'].values
    timestamp = df['datecompleted'].values
    activity_list[i][0] = steps
    activity_list[i][1] = goal
    activity_list[i][2] = timestamp

# calculate the average time difference in days between two activities -- feature 1
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    timestamp = activity_list[i][2]
    for j in range(len(timestamp)):
        ts = int(str(timestamp[j])[:-3])  # remove the last three digits from timestamp to make it specific to seconds
        timestamp[j] = ts
    
    # convert timestamp into dates
    dates = []
    for j in range(len(timestamp)):
        construct_time =time.gmtime(timestamp[j])  # turn timestamp into gmtime
        date = datetime.fromtimestamp(mktime(construct_time)).date()  # turn gmtime into datetime and remove time
        dates.append(date)
    
    if(steps.any()==0):  # 0 activity
        avg = 365  # the bigger avg is, the less frequent the activity is 
    else:
        # remove the date with 0 steps
        new_dates = []
        for j in range(len(steps)):
            if(steps[j]!=0):
                new_dates.append(dates[j])
        new_dates.sort()  # sort the dates

        if(len(new_dates)==1):  # only 1 activity
            avg = 100
        else:
            # calculate the time difference in days between two activities
            time_difference = []
            for j in range(len(new_dates)-1):
                difference = new_dates[j+1] - new_dates[j]
                difference = difference / timedelta(days=1)  # turn timedelta into float
                time_difference.append(difference)

            # calculate the average time difference
            sum_d = 0
            for j in range(len(time_difference)):
                sum_d = sum_d + time_difference[j]
            avg = round(sum_d / len(time_difference),2)
    features_list[i][0] = avg

# calculate the complete percentage -- feature 2
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    goal = activity_list[i][1]
    complete_num = 0
    for j in range(len(steps)):    
        if(steps[j]>=goal[j]):
            complete_num += 1
    complete_ratio = round(complete_num/len(goal),2)
    features_list[i][1] = complete_ratio

# calculate the average steps -- feature 3
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    steps_sum = 0
    if(steps.any()==0):
        steps_avg = 0
    else:
        for j in range(len(steps)):
            steps_sum += steps[j]
        steps_avg = round(steps_sum / len(steps),2)
    features_list[i][2] = steps_avg

features_list = np.array(features_list)

## apply kmeans
# Initializing KMeans
kmeans = KMeans(n_clusters=2)
# Fitting with inputs
# fit computes k-means clustering
kmeans = kmeans.fit(features_list)
# Predicting the clusters
# predict computes cluster centers and predict cluster index for each sample, returns index of the cluster each sample belong to
labels = kmeans.predict(features_list)
# Getting the cluster centers in coordinates
C = kmeans.cluster_centers_
visitor_clusterid = kmeans.labels_  # for each visitor(data point), get its cluster's id. 
print(visitor_clusterid)

## plot clustering results with shape 
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = visitor_clusterid
cluster0 = df_cm.loc[df_cm['cluster'] == 0]
cluster1 = df_cm.loc[df_cm['cluster'] == 1]
cluster2 = df_cm.loc[df_cm['cluster'] == 2]
cluster3 = df_cm.loc[df_cm['cluster'] == 3]
cluster4 = df_cm.loc[df_cm['cluster'] == 4]
cluster5 = df_cm.loc[df_cm['cluster'] == 5]

scatter0 = dict(
    mode = "markers",
    name = "Cluster 0",
    type = "scatter3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    marker = dict( size=2, color='orange')
)
scatter1 = dict(
    mode = "markers",
    name = "Cluster 1",
    type = "scatter3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    marker = dict( size=2, color='blue')
)
scatter2 = dict(
    mode = "markers",
    name = "Cluster 2",
    type = "scatter3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    marker = dict( size=2, color='red')
)
scatter3 = dict(
    mode = "markers",
    name = "Cluster 3",
    type = "scatter3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    marker = dict( size=2, color='green')
)
scatter4 = dict(
    mode = "markers",
    name = "Cluster 4",
    type = "scatter3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    marker = dict( size=2, color='grey')
)
scatter5 = dict(
    mode = "markers",
    name = "Cluster 5",
    type = "scatter3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    marker = dict( size=2, color='magenta')
)
cluster0 = dict(
    alphahull = 5,
    name = "Cluster 0",
    opacity = .1,
    type = "mesh3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    color='orange', showscale = True
)
cluster1 = dict(
    alphahull = 5,
    name = "Cluster 1",
    opacity = .1,
    type = "mesh3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    color='blue', showscale = True
)
cluster2 = dict(
    alphahull = 5,
    name = "Cluster 2",
    opacity = .1,
    type = "mesh3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    color='red', showscale = True
)
cluster3 = dict(
    alphahull = 5,
    name = "Cluster 3",
    opacity = .1,
    type = "mesh3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    color='green', showscale = True
)
cluster4 = dict(
    alphahull = 5,
    name = "Cluster 4",
    opacity = .1,
    type = "mesh3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    color='grey', showscale = True
)
cluster5 = dict(
    alphahull = 5,
    name = "Cluster 5",
    opacity = .1,
    type = "mesh3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    color='magenta', showscale = True
)
layout = dict(
    title = 'Activity data K-means clustering result ('+str(datetime.now().date())+")",
    scene = dict(
        xaxis = dict( title='avg_days',zeroline=True ),
        yaxis = dict( title='complete_ratio',zeroline=True ),
        zaxis = dict( title='avg_steps',zeroline=True ),
    )
)
fig = dict( data=[scatter0, scatter1, scatter2, scatter3, scatter4,scatter5, cluster0, cluster1, cluster2, cluster3, cluster4,cluster5], layout=layout )
plotly.offline.iplot(fig)

## use pca to reduce 3 features to 2 featrues for silhouette analysis
pca = PCA(n_components=2).fit(features_list)
X = pca.transform(features_list)  
#print(X)

## silhouette analysis
range_n_clusters = [2, 3, 4, 5, 6, 7]

for n_clusters in range_n_clusters:
    # Create a subplot with 1 row and 2 columns
    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(18, 7)

    # The 1st subplot is the silhouette plot
    # The silhouette coefficient can range from -1, 1 but in this example all
    # lie within [-0.1, 1]
    ax1.set_xlim([-0.1, 1])
    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters, to demarcate them clearly.
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])

    # Initialize the clusterer with n_clusters value and a random generator
    # seed of 10 for reproducibility.
    clusterer = KMeans(n_clusters=n_clusters, random_state=10)
    cluster_labels = clusterer.fit_predict(X)

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    silhouette_avg = silhouette_score(X, cluster_labels)
    print("For n_clusters =", n_clusters,
          "The average silhouette_score is :", silhouette_avg)

    # Compute the silhouette scores for each sample
    sample_silhouette_values = silhouette_samples(X, cluster_labels)

    y_lower = 10
    for i in range(n_clusters):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == i]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        color = cm.spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, ith_cluster_silhouette_values,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples

    ax1.set_title("The silhouette plot for the various clusters.")
    ax1.set_xlabel("The silhouette coefficient values")
    ax1.set_ylabel("Cluster label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

    # 2nd Plot showing the actual clusters formed
    colors = cm.spectral(cluster_labels.astype(float) / n_clusters)
    ax2.scatter(X[:, 0], X[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                c=colors, edgecolor='k')

    # Labeling the clusters
    centers = clusterer.cluster_centers_
    # Draw white circles at cluster centers
    ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                c="white", alpha=1, s=200, edgecolor='k')

    for i, c in enumerate(centers):
        ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                    s=50, edgecolor='k')

    ax2.set_title("The visualization of the clustered data.")
    ax2.set_xlabel("Feature space for the 1st feature")
    ax2.set_ylabel("Feature space for the 2nd feature")

    plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                  "with n_clusters = %d" % n_clusters),
                 fontsize=14, fontweight='bold')

    plt.show()
    
## heatmap of all users before normalization
x_labels = ['avg_days','complete_ratio','avg_steps']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (8,8))
plt.title("Activity data heatmap of all users")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g')# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## normalize features
features_list_transpose = np.array(features_list).transpose()
avg_days_list = features_list_transpose[0]
complete_ratio_list = features_list_transpose[1]
avg_steps_list = features_list_transpose[2]
avg_days_df = pd.DataFrame(avg_days_list)
complete_ratio_df = pd.DataFrame(complete_ratio_list)
avg_steps_df = pd.DataFrame(avg_steps_list)
avg_days_norm = (avg_days_df - avg_days_df.mean()) / (avg_days_df.max() - avg_days_df.min())
complete_ratio_norm = (complete_ratio_df - complete_ratio_df.mean()) / (complete_ratio_df.max() - complete_ratio_df.min())
avg_steps_norm = (avg_steps_df - avg_steps_df.mean()) / (avg_steps_df.max() - avg_steps_df.min())
features_list_transpose[0] = list(avg_days_norm[0].values)
features_list_transpose[1] = list(complete_ratio_norm[0].values)
features_list_transpose[2] = list(avg_steps_norm[0].values)

features_list = features_list_transpose.transpose()

# round to 2 decimals
for p in range(len(features_list)):
    for q in range(len(features_list[p])):
        features_list[p][q] = round(features_list[p][q],2)
        
## heatmap of all users after normalization
x_labels = ['avg_days','complete_ratio','avg_steps']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-1,vmax=1)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## heatmap for each cluster
df_cm['cluster'] = visitor_clusterid

df = df_cm.loc[df_cm['cluster'] == 0]
y_labels = df.index.values
plt.figure(figsize = (15,10))
plt.title("heatmap for cluster 0 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.set_context("talk")
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-1,vmax=1)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
