import pandas as pd
from pandas.plotting import scatter_matrix
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
import seaborn as sns
from datetime import datetime
from sklearn.metrics import silhouette_samples, silhouette_score
import matplotlib.cm as cm
from sklearn.decomposition import PCA
import numpy as np

## choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','amar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get csv data and put it into a matrix and do feature extraction. columns are features. rows are users
path = 'C:/Users/heyu/PycharmProjects/content data'
baseline_matrix = [['0']*5 for a in range(usersNum)]
# get five features from csv files
for i in range(usersNum):    
    filename = '/' + usersName[i] + '_baseline_data.csv'
    df = pd.read_csv(path+filename)
    age = df['answer'][0]
    gender = df['answer'][1]
    employment = df['answer'][2]
    pain_current = df['answer'][3]
    rmdq = df['answer'][4]
    baseline_matrix[i][0] = age
    baseline_matrix[i][1] = gender
    baseline_matrix[i][2] = employment
    baseline_matrix[i][3] = pain_current
    baseline_matrix[i][4] = rmdq

# feature extraction for age
for i in range(len(baseline_matrix)):
    age = int(baseline_matrix[i][0])
    if(age<25):
        baseline_matrix[i][0] = 0
    elif(25<=age<55):
        baseline_matrix[i][0] = 1
    elif(55<=age<65):
        baseline_matrix[i][0] = 2
    elif(65<=age):
        baseline_matrix[i][0] = 3
    else:
        baseline_matrix[i][0] = -1

# feature extraction for gender
for i in range(len(baseline_matrix)):
    gender = baseline_matrix[i][1]
    if(gender=='male'):
        baseline_matrix[i][1] = 0
    elif(gender=='female'):
        baseline_matrix[i][1] = 1
    else:
        baseline_matrix[i][1] = -1

# feature extraction for employment
for i in range(len(baseline_matrix)):
    employment = baseline_matrix[i][2]
    if(employment=='Full-time'):
        baseline_matrix[i][2] = 0
    elif(employment=='Part-time'):
        baseline_matrix[i][2] = 1
    elif(employment=='unemployed'):
        baseline_matrix[i][2] = 2
    elif(employment=='other'):
        baseline_matrix[i][2] = 3
    elif(employment=='full-time-housework'):
        baseline_matrix[i][2] = 4
    elif(employment=='retired'):
        baseline_matrix[i][2] = 5
    elif(employment=='military'):
        baseline_matrix[i][2] = 6
    else:
        baseline_matrix[i][2] = -1

# feature extraction for pain_current
for i in range(len(baseline_matrix)):
    pain_current = baseline_matrix[i][3]
    if(pain_current=='1week'):
        baseline_matrix[i][3] = 0
    elif(pain_current=='4weeks'):
        baseline_matrix[i][3] = 1
    elif(pain_current=='12weeks'):
        baseline_matrix[i][3] = 2
    elif(pain_current=='Above12weeks'):
        baseline_matrix[i][3] = 3
    else:
        baseline_matrix[i][3] = -1

# feature extraction for rmdq
for i in range(len(baseline_matrix)):
    rmdq = int(baseline_matrix[i][4])
    if(rmdq<5):
        baseline_matrix[i][4] = 0
    elif(5<=rmdq<10):
        baseline_matrix[i][4] = 1
    elif(10<=rmdq<15):
        baseline_matrix[i][4] = 2
    elif(15<=rmdq<20):
        baseline_matrix[i][4] = 3
    elif(20<=rmdq<25):
        baseline_matrix[i][4] = 4
    else:
        baseline_matrix[i][4] = -1

## use pca to reduce 3 features to 2 featrues for silhouette analysis
pca = PCA(n_components=2).fit(baseline_matrix)
X = pca.transform(baseline_matrix)
#print(X)

## silhouette analysis
range_n_clusters = [2, 3, 4, 5, 6]

for n_clusters in range_n_clusters:
    # Create a subplot with 1 row and 2 columns
    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(18, 7)

    # The 1st subplot is the silhouette plot
    # The silhouette coefficient can range from -1, 1 but in this example all
    # lie within [-0.1, 1]
    ax1.set_xlim([-0.1, 1])
    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters, to demarcate them clearly.
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])

    # Initialize the clusterer with n_clusters value and a random generator
    # seed of 10 for reproducibility.
    clusterer = KMeans(n_clusters=n_clusters, random_state=10)
    cluster_labels = clusterer.fit_predict(X)

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    silhouette_avg = silhouette_score(X, cluster_labels)
    print("For n_clusters =", n_clusters,
          "The average silhouette_score is :", silhouette_avg)

    # Compute the silhouette scores for each sample
    sample_silhouette_values = silhouette_samples(X, cluster_labels)

    y_lower = 10
    for i in range(n_clusters):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == i]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        color = cm.spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, ith_cluster_silhouette_values,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples

    ax1.set_title("The silhouette plot for the various clusters.")
    ax1.set_xlabel("The silhouette coefficient values")
    ax1.set_ylabel("Cluster label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

    # 2nd Plot showing the actual clusters formed
    colors = cm.spectral(cluster_labels.astype(float) / n_clusters)
    ax2.scatter(X[:, 0], X[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                c=colors, edgecolor='k')

    # Labeling the clusters
    centers = clusterer.cluster_centers_
    # Draw white circles at cluster centers
    ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                c="white", alpha=1, s=200, edgecolor='k')

    for i, c in enumerate(centers):
        ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                    s=50, edgecolor='k')

    ax2.set_title("The visualization of the clustered data.")
    ax2.set_xlabel("Feature space for the 1st feature")
    ax2.set_ylabel("Feature space for the 2nd feature")

    plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                  "with n_clusters = %d" % n_clusters),
                 fontsize=14, fontweight='bold')

    plt.show()

## heatmap of all users
x_labels = ['age','gender','employment','pain_current','rmdq']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(baseline_matrix, y_labels, x_labels)
plt.figure(figsize = (9,8))
plt.title("Baseline data heatmap of all users")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=0,vmax=4)# font size
plt.show()

## apply KMeans
baseline_df = pd.DataFrame(baseline_matrix)
# Initializing KMeans
kmeans = KMeans(n_clusters=3)
# Fitting with inputs
# fit computes k-means clustering
kmeans = kmeans.fit(baseline_matrix)
# Predicting the clusters
# predict computes cluster centers and predict cluster index for each sample, returns index of the cluster each sample belong to
labels = kmeans.predict(baseline_matrix)
# Getting the cluster centers in coordinates
C = kmeans.cluster_centers_  # the coordinates of cluster centers
print(C)
visitor_clusterid = kmeans.labels_  # for each visitor(data point), get its cluster's id. visitor 0 belongs to cluster 0
print(visitor_clusterid)

## add cluster id to the dataframe
df_cm['cluster'] = visitor_clusterid

## heatmap for each cluster
df = df_cm.loc[df_cm['cluster'] == 0]
# heatmap
x_labels = ['age','gender','employment','pain_current','rmdq']
y_labels = df.index.values
plt.figure(figsize = (9,5))
plt.title("Heatmap for cluster 0 of baseline data")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=0,vmax=4)# font size
plt.show()

