import pandas as pd
import numpy as np
from pandas.plotting import scatter_matrix
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
import seaborn as sns
import time
from time import mktime
from datetime import datetime
from datetime import timedelta
from mpl_toolkits.mplot3d import Axes3D
from itertools import groupby
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_samples, silhouette_score
import matplotlib.cm as cm
import plotly
plotly.offline.init_notebook_mode()

## choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','annelovise_no']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get exercise list
exerciseId_list = ['back_01_01', 'back_01_02', 'back_02_01', 'back_02_02', 'back_03_01', 'back_03_02', 'back_04_01', 'back_04_02', 'back_05_01', 'back_05_02', 'back_06_01', 'back_06_02', 'glut_01_01', 'glut_01_02', 'glut_02_01', 'glut_02_02', 'glut_03_01', 'glut_03_02', 'glut_04_01', 'glut_04_02', 'glut_05_01', 'glut_05_02', 'glut_06_01', 'glut_06_02', 'ab_01_01', 'ab_01_02', 'ab_02_01', 'ab_02_02', 'ab_03_01', 'ab_03_02', 'ab_03_03', 'ab_04_01', 'ab_04_02', 'ab_04_03', 'ab_05_01', 'ab_05_02', 'ab_06_01', 'ab_06_02', 'core_01_01', 'core_01_02', 'core_02_01', 'core_02_02', 'core_03_01', 'core_03_02', 'core_04_01', 'core_04_02', 'core_05_01', 'core_05_02', 'core_06_01', 'core_06_02', 'pain_01_01', 'pain_01_02', 'pain_02_01', 'pain_02_02', 'pain_02_03', 'pain_02_04', 'pain_03_01', 'pain_03_02', 'pain_03_03', 'pain_04_01', 'pain_04_02', 'flex_01_01', 'flex_01_02', 'flex_01_03', 'flex_01_04', 'flex_01_05', 'flex_01_06', 'flex_01_07', 'flex_01_08', 'flex_01_09']
print(len(exerciseId_list))

## get csv data and put it into a matrix with 3 features
path = 'C:/Users/heyu/PycharmProjects/content data'   
features_list = []
exercise_freq_matrix = []
avg_num_exercise_list = []
complete_ratio_list = []
for i in range(usersNum):
    # get completed data
    filename1 = '/' + usersName[i] + '_exercise_data.csv'
    df1 = pd.read_csv(path+filename1)
    exerciseIdCom = df1['exerciseid'].values
    timestampCom = df1['datecompleted'].values
    timestampCom.sort()
    tsCom_uni = []
    for j in range(len(timestampCom)):
        if(not(timestampCom[j] in tsCom_uni)):
            tsCom_uni.append(timestampCom[j])
    
    # get suggested data
    filename2 = '/' + usersName[i] + '_exercise_data_suggested.csv'
    df2 = pd.read_csv(path+filename2)
    exerciseIdSug = df2['exerciseid'].values
    timestampSug = df2['datesuggested'].values
    timestampSug.sort()
    tsSug_uni = []
    for j in(range(len(timestampSug))):
        if(not(timestampSug[j] in tsSug_uni)):
            tsSug_uni.append(timestampSug[j])
    
    # remove duplicated timestamp
    duplicate = 0
    for j in range(len(tsSug_uni)-1):
        if((tsSug_uni[j+1]-tsSug_uni[j])<1000):
            duplicate = tsSug_uni[j]
    df2 = df2[df2.datesuggested != duplicate]
    timestampSug = df2['datesuggested'].values
    timestampSug.sort()
    tsSug_uni = []
    for j in(range(len(timestampSug))):
        if(not(timestampSug[j] in tsSug_uni)):
            tsSug_uni.append(timestampSug[j])
    
    # use exercise ids in all sessions as feature 1, showing frequency information
    exercise_freq_list = [0]*len(exerciseId_list)
    for j in exerciseIdCom:
        index_j = exerciseId_list.index(j)
        exercise_freq_list[index_j] +=1
    exercise_freq_matrix.append(exercise_freq_list)

    # separate sessions
    sessionsCom = [list(g) for k, g in groupby(timestampCom)]
    sessionsSug = [list(g) for k, g in groupby(timestampSug)]
    
    # number of exercises in one session
    sessionCom_num = []
    for j in range(len(sessionsCom)):
        sessionCom_num.append(len(sessionsCom[j]))
    sessionSug_num = []
    for j in range(len(sessionsSug)):
        sessionSug_num.append(len(sessionsSug[j]))
    
    # calculate average number of exercises in one session (weeks with 0 exercises are not included) -- feature 2
    sum_num_exercise = 0
    for j in range(len(sessionCom_num)):
        sum_num_exercise = sum_num_exercise + sessionCom_num[j]
    avg_num_exercise = sum_num_exercise / len(sessionCom_num)
    avg_num_exercise_list.append(avg_num_exercise)
    
    # complete ratio of each week -- feature 3
    # calculate the expected timestamp
    tsExp = []
    for j in range(len(tsSug_uni)):
        ts_exp = tsSug_uni[j] + 604800000  # add 7 days
        tsExp.append(ts_exp)
    
    # get the suggested timestamp with exercises completed in that week
    tsSugCom = []
    for j in range(len(tsCom_uni)):
        for k in range(len(tsExp)):
            if(abs(tsCom_uni[j]-tsExp[k])<86400000):  # smaller than 1 day
                tsSugCom.append([tsCom_uni[j],tsSug_uni[k]])  # put the completed timestamp and the corresponding suggested timestamp into a pair
    
    # complete_ratio 
    complete_ratio = [0]*len(sessionSug_num)
    if(len(sessionSug_num)==len(sessionCom_num)):  # if every week there are exercises completed
        for j in range(len(sessionSug_num)):
            complete_ratio[j] = sessionCom_num[j] / sessionSug_num[j]  # for weeks with exercises cpmpleted
    else:  # if there is a week or weeks with no exercise completed
        for j in tsSugCom:
            index_sug = tsSug_uni.index(j[1])
            index_com = tsCom_uni.index(j[0])
            complete_ratio[index_sug] = sessionCom_num[index_com] / sessionSug_num[index_sug]  # put non_zero complete ratio to the weeks with exercises completed
    complete_ratio_list.append(complete_ratio)

# apply pca to exercise_freq_matrix
pca = PCA(n_components=1)
principalComponents0 = pca.fit_transform(exercise_freq_matrix)
exercise_freq_users = pd.DataFrame(data = principalComponents0)
exercise_freq_users = exercise_freq_users[0].values.tolist()

# pad 0s to complete ratio list to make matrix
pad = len(max(complete_ratio_list,key=len))
complete_ratio_list = np.array([i + [0]*(pad-len(i)) for i in complete_ratio_list])
# apply pca on complete_ratio_matrix
pca = PCA(n_components=1)
principalComponents1 = pca.fit_transform(complete_ratio_list)
complete_ratio_users = pd.DataFrame(data = principalComponents1)
complete_ratio_users = complete_ratio_users[0].values.tolist()

features_list.append(exercise_freq_users)
features_list.append(avg_num_exercise_list)
features_list.append(complete_ratio_users)

# transpose the features list
features_list = np.transpose(features_list)

# round each number to 2 decimals
for i in range(len(features_list)):
    for j in range(len(features_list[i])):
        features_list[i][j] = round(features_list[i][j],2)

## use pca to reduce 3 features to 2 featrues for silhouette analysis
pca = PCA(n_components=2).fit(features_list)
X = pca.transform(features_list)  
#print(X)

## silhouette analysis
range_n_clusters = [2, 3, 4, 5, 6]

for n_clusters in range_n_clusters:
    # Create a subplot with 1 row and 2 columns
    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(18, 7)

    # The 1st subplot is the silhouette plot
    # The silhouette coefficient can range from -1, 1 but in this example all
    # lie within [-0.1, 1]
    ax1.set_xlim([-0.1, 1])
    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters, to demarcate them clearly.
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])

    # Initialize the clusterer with n_clusters value and a random generator
    # seed of 10 for reproducibility.
    clusterer = KMeans(n_clusters=n_clusters, random_state=10)
    cluster_labels = clusterer.fit_predict(X)

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    silhouette_avg = silhouette_score(X, cluster_labels)
    print("For n_clusters =", n_clusters,
          "The average silhouette_score is :", silhouette_avg)

    # Compute the silhouette scores for each sample
    sample_silhouette_values = silhouette_samples(X, cluster_labels)

    y_lower = 10
    for i in range(n_clusters):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == i]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        color = cm.spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, ith_cluster_silhouette_values,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples

    ax1.set_title("The silhouette plot for the various clusters.")
    ax1.set_xlabel("The silhouette coefficient values")
    ax1.set_ylabel("Cluster label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

    # 2nd Plot showing the actual clusters formed
    colors = cm.spectral(cluster_labels.astype(float) / n_clusters)
    ax2.scatter(X[:, 0], X[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                c=colors, edgecolor='k')

    # Labeling the clusters
    centers = clusterer.cluster_centers_
    # Draw white circles at cluster centers
    ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                c="white", alpha=1, s=200, edgecolor='k')

    for i, c in enumerate(centers):
        ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                    s=50, edgecolor='k')

    ax2.set_title("The visualization of the clustered data.")
    ax2.set_xlabel("Feature space for the 1st feature")
    ax2.set_ylabel("Feature space for the 2nd feature")

    plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                  "with n_clusters = %d" % n_clusters),
                 fontsize=14, fontweight='bold')

    plt.show()
    
## apply kmeans
# Initializing KMeans
kmeans = KMeans(n_clusters=2)
# Fitting with inputs
# fit computes k-means clustering
kmeans = kmeans.fit(features_list)
# Predicting the clusters
# predict computes cluster centers and predict cluster index for each sample, returns index of the cluster each sample belong to
labels = kmeans.predict(features_list)
# Getting the cluster centers in coordinates
C = kmeans.cluster_centers_
visitor_clusterid = kmeans.labels_  # for each visitor(data point), get its cluster's id. 
print(visitor_clusterid)

## plot clustering result with shape
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = visitor_clusterid
cluster0 = df_cm.loc[df_cm['cluster'] == 0]
cluster1 = df_cm.loc[df_cm['cluster'] == 1]
cluster2 = df_cm.loc[df_cm['cluster'] == 2]
cluster3 = df_cm.loc[df_cm['cluster'] == 3]
cluster4 = df_cm.loc[df_cm['cluster'] == 4]
cluster5 = df_cm.loc[df_cm['cluster'] == 5]

scatter0 = dict(
    mode = "markers",
    name = "Cluster 0",
    type = "scatter3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    marker = dict( size=2, color='green')
)
scatter1 = dict(
    mode = "markers",
    name = "Cluster 1",
    type = "scatter3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    marker = dict( size=2, color='blue')
)
scatter2 = dict(
    mode = "markers",
    name = "Cluster 2",
    type = "scatter3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    marker = dict( size=2, color='red')
)
scatter3 = dict(
    mode = "markers",
    name = "Cluster 3",
    type = "scatter3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    marker = dict( size=2, color='magenta')
)
scatter4 = dict(
    mode = "markers",
    name = "Cluster 4",
    type = "scatter3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    marker = dict( size=2, color='grey')
)
scatter5 = dict(
    mode = "markers",
    name = "Cluster 5",
    type = "scatter3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    marker = dict( size=2, color='yellow')
)
cluster0 = dict(
    alphahull = 5,
    name = "Cluster 0",
    opacity = .1,
    type = "mesh3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    color='green', showscale = True
)
cluster1 = dict(
    alphahull = 5,
    name = "Cluster 1",
    opacity = .1,
    type = "mesh3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    color='blue', showscale = True
)
cluster2 = dict(
    alphahull = 5,
    name = "Cluster 2",
    opacity = .1,
    type = "mesh3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    color='red', showscale = True
)
cluster3 = dict(
    alphahull = 5,
    name = "Cluster 3",
    opacity = .1,
    type = "mesh3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    color='magenta', showscale = True
)
cluster4 = dict(
    alphahull = 5,
    name = "Cluster 4",
    opacity = .1,
    type = "mesh3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    color='grey', showscale = True
)
cluster5 = dict(
    alphahull = 5,
    name = "Cluster 5",
    opacity = .1,
    type = "mesh3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    color='yellow', showscale = True
)
layout = dict(
    title = 'Exercise data K-means clustering result ('+str(datetime.now().date())+")",
    scene = dict(
        xaxis = dict( title='exe_freq',zeroline=True ),
        yaxis = dict( title='avg_num_exe_per_session',zeroline=True ),
        zaxis = dict( title='complete_ratio',zeroline=True ),
    )
)
fig = dict( data=[scatter0, scatter1, scatter2,scatter3,scatter4, scatter5, cluster0, cluster1, cluster2,cluster3,cluster4,cluster5], layout=layout )
plotly.offline.iplot(fig)

## heatmap of all users before normalization
x_labels = ['exe_freq','avg_num_exe_per_session','complete_ratio']
#y_labels = ['agnar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (9,8))
plt.title("Exercise data heatmap of all users")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g')# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## normalize features
exe_freq_df = pd.DataFrame(exercise_freq_users)
avg_num_exe_df = pd.DataFrame(avg_num_exercise_list)
complete_ratio_df = pd.DataFrame(complete_ratio_users)

exe_freq_norm = (exe_freq_df - exe_freq_df.mean()) / (exe_freq_df.max() - exe_freq_df.min())
avg_num_exe_norm = (avg_num_exe_df - avg_num_exe_df.mean()) / (avg_num_exe_df.max() - avg_num_exe_df.min())
complete_ratio_norm = (complete_ratio_df - complete_ratio_df.mean()) / (complete_ratio_df.max() - complete_ratio_df.min())

exercise_freq_users_norm = list(exe_freq_norm[0].values)
avg_num_exercise_list_norm = list(avg_num_exe_norm[0].values)
complete_ratio_users_norm = list(complete_ratio_norm[0].values)

features_list_norm = []
features_list_norm.append(exercise_freq_users_norm)
features_list_norm.append(avg_num_exercise_list_norm)
features_list_norm.append(complete_ratio_users_norm)

# round each number to 2 decimals
for i in range(len(features_list_norm)):
    for j in range(len(features_list_norm[i])):
        features_list_norm[i][j] = round(features_list_norm[i][j],2)
        
# transpose the features list
features_list = np.transpose(features_list_norm)

## heatmap of all users after normalization
x_labels = ['exe_freq','avg_num_exe_per_session','complete_ratio']
#y_labels = ['agnar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-0.6,vmax=0.6)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## heatmap for each cluster
df_cm['cluster'] = visitor_clusterid

df = df_cm.loc[df_cm['cluster'] == 1]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 0 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-0.6,vmax=0.6)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
