import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
import json
import requests
from pandas.io.json import json_normalize
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
import scipy.spatial.distance as ssd
import random
from collections import Counter
import matplotlib.patches as mpatches

## get backup sessions data
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "sessions" + ".txt"
sessionList = []
with open(path+filename,"r") as file:
    session = eval(file.readline())
    sessionList.extend(session)
    
## normalize time spent
flow_df = pd.DataFrame(sessionList[0])
time_df = pd.DataFrame(sessionList[1])
flow_pca2_norm = (flow_df - flow_df.mean()) / (flow_df.max() - flow_df.min())
time_pca_norm = (time_df - time_df.mean()) / (time_df.max() - time_df.min())
features_list = []
features_list.append(list(flow_pca2_norm[0].values))
features_list.append(list(time_pca_norm[0].values))
features_list = np.transpose(features_list)

## apply kmedoids
def cluster(distances, k=3):

    m = distances.shape[0] # number of points

    # Pick k random medoids.
    curr_medoids = np.array([-1]*k)
    while not len(np.unique(curr_medoids)) == k:
        curr_medoids = np.array([random.randint(0, m - 1) for _ in range(k)])
    old_medoids = np.array([-1]*k) # Doesn't matter what we initialize these to.
    new_medoids = np.array([-1]*k)
   
    # Until the medoids stop updating, do the following:
    while not ((old_medoids == curr_medoids).all()):
        # Assign each point to cluster with closest medoid.
        clusters = assign_points_to_clusters(curr_medoids, distances)

        # Update cluster medoids to be lowest cost point. 
        for curr_medoid in curr_medoids:
            cluster = np.where(clusters == curr_medoid)[0]
            new_medoids[curr_medoids == curr_medoid] = compute_new_medoid(cluster, distances)

        old_medoids[:] = curr_medoids[:]
        curr_medoids[:] = new_medoids[:]

    return clusters, curr_medoids

def assign_points_to_clusters(medoids, distances):
    distances_to_medoids = distances[:,medoids]
    clusters = medoids[np.argmin(distances_to_medoids, axis=1)]
    clusters[medoids] = medoids
    return clusters

def compute_new_medoid(cluster, distances):
    mask = np.ones(distances.shape)
    mask[np.ix_(cluster,cluster)] = 0.
    cluster_distances = np.ma.masked_array(data=distances, mask=mask, fill_value=10e9)
    costs = cluster_distances.sum(axis=1)
    return costs.argmin(axis=0, fill_value=10e9)

distanceMatrix = ssd.squareform(ssd.pdist(features_list, 'cosine'))
clusters = 4
clusters, medoids = cluster(distanceMatrix,clusters)
print(clusters)
print(medoids)

## name the clusters by ids instead of medoids
for i in range(len(clusters)):
    clusters[i] = list(medoids).index(clusters[i])
print(clusters)

## add the clustering result to the data
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = clusters

## plot clustering result
colors = []
for i in range(len(clusters)):
    if(clusters[i]==0):
        color = 'purple'
    elif(clusters[i]==1):
        color = 'deepskyblue'
    elif(clusters[i]==2):
        color = 'green'
    elif(clusters[i]==3):
        color = 'red'
    elif(clusters[i]==4):
        color = 'magenta'
    elif(clusters[i]==5):
        color = 'blue'
    elif(clusters[i]==6):
        color = 'grey'
    elif(clusters[i]==7):
        color = 'gold'
    colors.append(color)

plt.title("Sessions data K-medoids clustering result ("+str(datetime.now().date())+")")
plt.xlabel('userflow')
plt.ylabel('timespent')
patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend
plt.scatter(features_list[:, 0], features_list[:, 1], c=colors)

## get total visits of each user
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "totalVisits" + ".txt"
user_totalVisits = []
with open(path+filename,"r") as file:
    totalVisits = eval(file.readline())
    user_totalVisits.extend(totalVisits)
print(user_totalVisits)

## add cluster ids
df_c = pd.DataFrame(features_list)
df_c['cluster'] = clusters

## get clustering results of each user's sessions
cluster_user = []
starting_index = 0
print(user_totalVisits)
for i in range(len(user_totalVisits)):
    totalVisits = user_totalVisits[i]
    df = df_c.iloc[starting_index:starting_index+totalVisits]
    cluster = list(df['cluster'].values)
    cluster_user.append(cluster)
    starting_index = starting_index + totalVisits
print(len(cluster_user))
print(cluster_user)

## plot user-session-cluster
#users = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
users = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

plt.figure(figsize=(15,10))
plt.title("Users' sessions clustering result ("+str(datetime.now().date())+")")
plt.xlabel('user')
plt.ylabel('session')
plt.yticks(np.arange(0,10,step=1))

patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend
for i in range(len(cluster_user)):
    for j in range(len(cluster_user[i])):
        #plt.scatter(users[i],cluster_user[i][j])
        if(cluster_user[i][j]==0):
            plt.scatter(users[i],j,c='purple')
        elif(cluster_user[i][j]==1):
            plt.scatter(users[i],j,c='deepskyblue')
        elif(cluster_user[i][j]==2):
            plt.scatter(users[i],j,c='green')
        elif(cluster_user[i][j]==3):
            plt.scatter(users[i],j,c='red')
            
## show each user's cluster distribution
df_u = pd.DataFrame(cluster_user)

# show the representative cluster of each user
mode_user = []
for i in range(len(cluster_user)):
    data = Counter(cluster_user[i])
    mode = data.most_common(1)
    mode_user.append(mode)
df_u['mode'] = pd.DataFrame(mode_user)

df_u['user'] = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
#df_u['user'] = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

df_u            
