import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
import scipy.spatial.distance as ssd
import random

## concatenate all matrices
path = "C:/Users/heyu/PycharmProjects/backup/big matrix records/"
oldBigMatrix = []
for i in range(4):  # 4 files
    filename = "bigMatrix_studyTestUsers_" + str(i) + ".txt"
    with open(path+filename,"r") as file:
        matrix = eval(file.readline())
        oldBigMatrix.extend(matrix)
print(np.array(oldBigMatrix).shape)

# remove duplicated sessions
bigMatrix = []
for i in oldBigMatrix:
    if(i not in bigMatrix):
        bigMatrix.append(i)
print(np.array(bigMatrix).shape)

visitors = len(bigMatrix)
sessions = len(bigMatrix[0])
print(visitors,sessions)

## use pca to reduce dimensions
pca_3d = []
for i in range(visitors):
    pca_list = []
    for j in range(sessions):
        userflow = bigMatrix[i][j]
        pca = PCA(n_components=1)
        principalComponents = pca.fit_transform(userflow)
        df = pd.DataFrame(data = principalComponents)
        df = df[0].values.tolist()
        
        pca_list.append(df)
    pca_3d.append(pca_list)  # till here every 33-list in pca_3d is corresponding to the userflow matrix

# apply pca again
pca1_list =[]
for m in range(visitors):
    pca1 = PCA(n_components=1)
    principalComponents1 = pca1.fit_transform(pca_3d[m])
    df1 = pd.DataFrame(data = principalComponents1)
    df1 = df1[0].values.tolist()
    pca1_list.append(df1)

# round numbers to 2 decimals. 
# the cluster results of 5 decimals, 3 decimals and before rounding are the same
# the heatmap is created by around 2 decimals
for p in range(len(pca1_list)):
    for q in range(len(pca1_list[p])):
        pca1_list[p][q] = round(pca1_list[p][q],2)
#print(pca1_list)
pca1_list = np.array(pca1_list)

## apply kmedoids
def cluster(distances, k=3):

    m = distances.shape[0] # number of points

    # Pick k random medoids.
    curr_medoids = np.array([-1]*k)
    while not len(np.unique(curr_medoids)) == k:
        curr_medoids = np.array([random.randint(0, m - 1) for _ in range(k)])
    old_medoids = np.array([-1]*k) # Doesn't matter what we initialize these to.
    new_medoids = np.array([-1]*k)
   
    # Until the medoids stop updating, do the following:
    while not ((old_medoids == curr_medoids).all()):
        # Assign each point to cluster with closest medoid.
        clusters = assign_points_to_clusters(curr_medoids, distances)

        # Update cluster medoids to be lowest cost point. 
        for curr_medoid in curr_medoids:
            cluster = np.where(clusters == curr_medoid)[0]
            new_medoids[curr_medoids == curr_medoid] = compute_new_medoid(cluster, distances)

        old_medoids[:] = curr_medoids[:]
        curr_medoids[:] = new_medoids[:]

    return clusters, curr_medoids

def assign_points_to_clusters(medoids, distances):
    distances_to_medoids = distances[:,medoids]
    clusters = medoids[np.argmin(distances_to_medoids, axis=1)]
    clusters[medoids] = medoids
    return clusters

def compute_new_medoid(cluster, distances):
    mask = np.ones(distances.shape)
    mask[np.ix_(cluster,cluster)] = 0.
    cluster_distances = np.ma.masked_array(data=distances, mask=mask, fill_value=10e9)
    costs = cluster_distances.sum(axis=1)
    return costs.argmin(axis=0, fill_value=10e9)

distanceMatrix = ssd.squareform(ssd.pdist(pca1_list, 'cosine'))
clusters = 4
clusters, medoids = cluster(distanceMatrix,clusters)
print(clusters)
print(medoids)

## heatmap of all users
x_labels = [str('s'+ str(i)) for i in range(sessions)]
y_labels = [str('u'+ str(i)) for i in range(visitors)]
df_cm = pd.DataFrame(pca1_list,y_labels, x_labels)
plt.figure(figsize=(10,35))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g')# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()

## heatmap for each cluster
df_cm['cluster'] = clusters

df = df_cm.loc[df_cm['cluster'] == medoids[3]]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 3 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-4,vmax=16)# font size
plt.xlabel('session')
plt.ylabel("user")
plt.show()
