import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
import scipy.spatial.distance as ssd
import random
import time
from time import mktime
from datetime import timedelta
import plotly
plotly.offline.init_notebook_mode()

##choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','amar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get csv data and put it into a 3D list -- features_list, which is used in the clustering method
path = 'C:/Users/heyu/PycharmProjects/content data'   
activity_list = [[0]*3 for a in range(usersNum)]
features_list = [[0]*3 for a in range(usersNum)]

# get three types of activity data to construct features
for i in range(usersNum):    
    filename = '/' + usersName[i] + '_activity_data.csv'
    df = pd.read_csv(path+filename)
    steps = df['completed'].values
    goal = df['suggested'].values
    timestamp = df['datecompleted'].values
    activity_list[i][0] = steps
    activity_list[i][1] = goal
    activity_list[i][2] = timestamp

# calculate the average time difference in days between two activities -- feature 1
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    timestamp = activity_list[i][2]
    for j in range(len(timestamp)):
        ts = int(str(timestamp[j])[:-3])  # remove the last three digits from timestamp to make it specific to seconds
        timestamp[j] = ts
    
    # convert timestamp into dates
    dates = []
    for j in range(len(timestamp)):
        construct_time =time.gmtime(timestamp[j])  # turn timestamp into gmtime
        date = datetime.fromtimestamp(mktime(construct_time)).date()  # turn gmtime into datetime and remove time
        dates.append(date)
    
    if(steps.any()==0):  # 0 activity
        avg = 365  # the bigger avg is, the less frequent the activity is 
    else:
        # remove the date with 0 steps
        new_dates = []
        for j in range(len(steps)):
            if(steps[j]!=0):
                new_dates.append(dates[j])
        new_dates.sort()  # sort the dates

        if(len(new_dates)==1):  # only 1 activity
            avg = 100
        else:
            # calculate the time difference in days between two activities
            time_difference = []
            for j in range(len(new_dates)-1):
                difference = new_dates[j+1] - new_dates[j]
                difference = difference / timedelta(days=1)  # turn timedelta into float
                time_difference.append(difference)

            # calculate the average time difference
            sum_d = 0
            for j in range(len(time_difference)):
                sum_d = sum_d + time_difference[j]
            avg = round(sum_d / len(time_difference),2)
    features_list[i][0] = avg

# calculate the complete percentage -- feature 2
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    goal = activity_list[i][1]
    complete_num = 0
    for j in range(len(steps)):    
        if(steps[j]>=goal[j]):
            complete_num += 1
    complete_ratio = round(complete_num/len(goal),2)
    features_list[i][1] = complete_ratio

# calculate the average steps -- feature 3
for i in range(len(activity_list)):
    steps = activity_list[i][0]
    steps_sum = 0
    if(steps.any()==0):
        steps_avg = 0
    else:
        for j in range(len(steps)):
            steps_sum += steps[j]
        steps_avg = round(steps_sum / len(steps),2)
    features_list[i][2] = steps_avg

## apply kmedoids
def cluster(distances, k=3):

    m = distances.shape[0] # number of points

    # Pick k random medoids.
    curr_medoids = np.array([-1]*k)
    while not len(np.unique(curr_medoids)) == k:
        curr_medoids = np.array([random.randint(0, m - 1) for _ in range(k)])
    old_medoids = np.array([-1]*k) # Doesn't matter what we initialize these to.
    new_medoids = np.array([-1]*k)
   
    # Until the medoids stop updating, do the following:
    while not ((old_medoids == curr_medoids).all()):
        # Assign each point to cluster with closest medoid.
        clusters = assign_points_to_clusters(curr_medoids, distances)

        # Update cluster medoids to be lowest cost point. 
        for curr_medoid in curr_medoids:
            cluster = np.where(clusters == curr_medoid)[0]
            new_medoids[curr_medoids == curr_medoid] = compute_new_medoid(cluster, distances)

        old_medoids[:] = curr_medoids[:]
        curr_medoids[:] = new_medoids[:]

    return clusters, curr_medoids

def assign_points_to_clusters(medoids, distances):
    distances_to_medoids = distances[:,medoids]
    clusters = medoids[np.argmin(distances_to_medoids, axis=1)]  # np.argmin returns the indeices of the min values along an axis
    clusters[medoids] = medoids
    return clusters

def compute_new_medoid(cluster, distances):
    mask = np.ones(distances.shape)
    mask[np.ix_(cluster,cluster)] = 0.  # ix_ construct an open mesh from multiple sequences
    cluster_distances = np.ma.masked_array(data=distances, mask=mask, fill_value=10e9)  # to manipulate numerical arrays with missing data. mask is boolean in which True means the elment is invalid. the invalid data will be replaced by fill_value
    costs = cluster_distances.sum(axis=1)
    return costs.argmin(axis=0, fill_value=10e9)

distanceMatrix = ssd.squareform(ssd.pdist(features_list, 'cosine'))  # ssd.pdist returns the pairwise cosine distances between observations in n-dimensional space. squareform converts a vector-form distance vector to a square-form distance matrix
clusters = 6
clusters, medoids = cluster(distanceMatrix,clusters)
print(clusters)
print(medoids)

## add the clustering result to the data
df_cm = pd.DataFrame(features_list)
df_cm['cluster'] = clusters

## plot clustering result with shape
cluster0 = df_cm.loc[df_cm['cluster'] == medoids[0]]
cluster1 = df_cm.loc[df_cm['cluster'] == medoids[1]]
cluster2 = df_cm.loc[df_cm['cluster'] == medoids[2]]
cluster3 = df_cm.loc[df_cm['cluster'] == medoids[3]]
cluster4 = df_cm.loc[df_cm['cluster'] == medoids[4]]
cluster5 = df_cm.loc[df_cm['cluster'] == medoids[5]]

scatter0 = dict(
    mode = "markers",
    name = "Cluster 0",
    type = "scatter3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    marker = dict( size=2, color='orange')
)
scatter1 = dict(
    mode = "markers",
    name = "Cluster 1",
    type = "scatter3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    marker = dict( size=2, color='blue')
)
scatter2 = dict(
    mode = "markers",
    name = "Cluster 2",
    type = "scatter3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    marker = dict( size=2, color='red')
)
scatter3 = dict(
    mode = "markers",
    name = "Cluster 3",
    type = "scatter3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    marker = dict( size=2, color='green')
)
scatter4 = dict(
    mode = "markers",
    name = "Cluster 4",
    type = "scatter3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    marker = dict( size=2, color='grey')
)
scatter5 = dict(
    mode = "markers",
    name = "Cluster 5",
    type = "scatter3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    marker = dict( size=2, color='magenta')
)
cluster0 = dict(
    alphahull = 5,
    name = "Cluster 0",
    opacity = .1,
    type = "mesh3d",    
    x = cluster0.as_matrix()[:,0], y = cluster0.as_matrix()[:,1], z = cluster0.as_matrix()[:,2],
    color='orange', showscale = True
)
cluster1 = dict(
    alphahull = 5,
    name = "Cluster 1",
    opacity = .1,
    type = "mesh3d",    
    x = cluster1.as_matrix()[:,0], y = cluster1.as_matrix()[:,1], z = cluster1.as_matrix()[:,2],
    color='blue', showscale = True
)
cluster2 = dict(
    alphahull = 5,
    name = "Cluster 2",
    opacity = .1,
    type = "mesh3d",    
    x = cluster2.as_matrix()[:,0], y = cluster2.as_matrix()[:,1], z = cluster2.as_matrix()[:,2],
    color='red', showscale = True
)
cluster3 = dict(
    alphahull = 5,
    name = "Cluster 3",
    opacity = .1,
    type = "mesh3d",    
    x = cluster3.as_matrix()[:,0], y = cluster3.as_matrix()[:,1], z = cluster3.as_matrix()[:,2],
    color='green', showscale = True
)
cluster4 = dict(
    alphahull = 5,
    name = "Cluster 4",
    opacity = .1,
    type = "mesh3d",    
    x = cluster4.as_matrix()[:,0], y = cluster4.as_matrix()[:,1], z = cluster4.as_matrix()[:,2],
    color='grey', showscale = True
)
cluster5 = dict(
    alphahull = 5,
    name = "Cluster 5",
    opacity = .1,
    type = "mesh3d",    
    x = cluster5.as_matrix()[:,0], y = cluster5.as_matrix()[:,1], z = cluster5.as_matrix()[:,2],
    color='magenta', showscale = True
)
layout = dict(
    title = 'Activity data K-means clustering result ('+str(datetime.now().date())+")",
    scene = dict(
        xaxis = dict( title='avg_days',zeroline=True ),
        yaxis = dict( title='complete_ratio',zeroline=True ),
        zaxis = dict( title='avg_steps',zeroline=True ),
    )
)
fig = dict( data=[scatter0, scatter1, scatter2, scatter3, scatter4,scatter5, cluster0, cluster1, cluster2, cluster3, cluster4,cluster5], layout=layout )
plotly.offline.iplot(fig)

## normalize features
features_list_transpose = np.array(features_list).transpose()

avg_days_list = features_list_transpose[0]
complete_ratio_list = features_list_transpose[1]
avg_steps_list = features_list_transpose[2]

avg_days_df = pd.DataFrame(avg_days_list)
complete_ratio_df = pd.DataFrame(complete_ratio_list)
avg_steps_df = pd.DataFrame(avg_steps_list)
avg_days_norm = (avg_days_df - avg_days_df.mean()) / (avg_days_df.max() - avg_days_df.min())
complete_ratio_norm = (complete_ratio_df - complete_ratio_df.mean()) / (complete_ratio_df.max() - complete_ratio_df.min())
avg_steps_norm = (avg_steps_df - avg_steps_df.mean()) / (avg_steps_df.max() - avg_steps_df.min())
features_list_transpose[0] = list(avg_days_norm[0].values)
features_list_transpose[1] = list(complete_ratio_norm[0].values)
features_list_transpose[2] = list(avg_steps_norm[0].values)

features_list = features_list_transpose.transpose()

# round to 2 decimals
for p in range(len(features_list)):
    for q in range(len(features_list[p])):
        features_list[p][q] = round(features_list[p][q],2)
        
## heatmap of all users after normalization
x_labels = ['avg_days','complete_ratio','avg_steps']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(features_list, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-1,vmax=1)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## heatmap for each cluster
df_cm['cluster'] = clusters
df = df_cm.loc[df_cm['cluster'] == medoids[3]]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 3("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=-1,vmax=1)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
