import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
import scipy.spatial.distance as ssd
import random

## choose users to study
studyUsersName = ['tfs_aud_02','tfs_berit_05','tfs_jan_01','tfs_jorgen_07','tfs_karen_04','tfs_lill_03','tfs_line_08','tfs_oystein_09','tsf_tone_10']
testUsersName = ['ilya','sadiq','stewart','td','louise','malene','paul_mail_test_no','kerstin','keba_en','karen','charlotte','nirmalie','agnar','amar','annelovise_no','Tom_TR_Test']
usersName = studyUsersName + testUsersName
usersName.sort()             
print(usersName)
usersNum = len(usersName)
print(usersNum)

## get csv data and put it into a matrix and do feature extraction. columns are features. rows are users
path = 'C:/Users/heyu/PycharmProjects/content data'
baseline_matrix = [['0']*5 for a in range(usersNum)]
# get five features from csv files
for i in range(usersNum):    
    filename = '/' + usersName[i] + '_baseline_data.csv'
    df = pd.read_csv(path+filename)
    age = df['answer'][0]
    gender = df['answer'][1]
    employment = df['answer'][2]
    pain_current = df['answer'][3]
    rmdq = df['answer'][4]
    baseline_matrix[i][0] = age
    baseline_matrix[i][1] = gender
    baseline_matrix[i][2] = employment
    baseline_matrix[i][3] = pain_current
    baseline_matrix[i][4] = rmdq

# feature extraction for age
for i in range(len(baseline_matrix)):
    age = int(baseline_matrix[i][0])
    if(age<25):
        baseline_matrix[i][0] = 0
    elif(25<=age<55):
        baseline_matrix[i][0] = 1
    elif(55<=age<65):
        baseline_matrix[i][0] = 2
    elif(65<=age):
        baseline_matrix[i][0] = 3
    else:
        baseline_matrix[i][0] = -1

# feature extraction for gender
for i in range(len(baseline_matrix)):
    gender = baseline_matrix[i][1]
    if(gender=='male'):
        baseline_matrix[i][1] = 0
    elif(gender=='female'):
        baseline_matrix[i][1] = 1
    else:
        baseline_matrix[i][1] = -1

# feature extraction for employment
for i in range(len(baseline_matrix)):
    employment = baseline_matrix[i][2]
    if(employment=='Full-time'):
        baseline_matrix[i][2] = 0
    elif(employment=='Part-time'):
        baseline_matrix[i][2] = 1
    elif(employment=='unemployed'):
        baseline_matrix[i][2] = 2
    elif(employment=='other'):
        baseline_matrix[i][2] = 3
    elif(employment=='full-time-housework'):
        baseline_matrix[i][2] = 4
    elif(employment=='retired'):
        baseline_matrix[i][2] = 5
    elif(employment=='military'):
        baseline_matrix[i][2] = 6
    else:
        baseline_matrix[i][2] = -1

# feature extraction for pain_current
for i in range(len(baseline_matrix)):
    pain_current = baseline_matrix[i][3]
    if(pain_current=='1week'):
        baseline_matrix[i][3] = 0
    elif(pain_current=='4weeks'):
        baseline_matrix[i][3] = 1
    elif(pain_current=='12weeks'):
        baseline_matrix[i][3] = 2
    elif(pain_current=='Above12weeks'):
        baseline_matrix[i][3] = 3
    else:
        baseline_matrix[i][3] = -1

# feature extraction for rmdq
for i in range(len(baseline_matrix)):
    rmdq = int(baseline_matrix[i][4])
    if(rmdq<5):
        baseline_matrix[i][4] = 0
    elif(5<=rmdq<10):
        baseline_matrix[i][4] = 1
    elif(10<=rmdq<15):
        baseline_matrix[i][4] = 2
    elif(15<=rmdq<20):
        baseline_matrix[i][4] = 3
    elif(20<=rmdq<25):
        baseline_matrix[i][4] = 4
    else:
        baseline_matrix[i][4] = -1

features_list = np.array(baseline_matrix)

## apply kmedoids
def cluster(distances, k=3):

    m = distances.shape[0] # number of points

    # Pick k random medoids.
    curr_medoids = np.array([-1]*k)
    while not len(np.unique(curr_medoids)) == k:
        curr_medoids = np.array([random.randint(0, m - 1) for _ in range(k)])
    old_medoids = np.array([-1]*k) # Doesn't matter what we initialize these to.
    new_medoids = np.array([-1]*k)
   
    # Until the medoids stop updating, do the following:
    while not ((old_medoids == curr_medoids).all()):
        # Assign each point to cluster with closest medoid.
        clusters = assign_points_to_clusters(curr_medoids, distances)

        # Update cluster medoids to be lowest cost point. 
        for curr_medoid in curr_medoids:
            cluster = np.where(clusters == curr_medoid)[0]
            new_medoids[curr_medoids == curr_medoid] = compute_new_medoid(cluster, distances)

        old_medoids[:] = curr_medoids[:]
        curr_medoids[:] = new_medoids[:]

    return clusters, curr_medoids

def assign_points_to_clusters(medoids, distances):
    distances_to_medoids = distances[:,medoids]
    clusters = medoids[np.argmin(distances_to_medoids, axis=1)]
    clusters[medoids] = medoids
    return clusters

def compute_new_medoid(cluster, distances):
    mask = np.ones(distances.shape)
    mask[np.ix_(cluster,cluster)] = 0.
    cluster_distances = np.ma.masked_array(data=distances, mask=mask, fill_value=10e9)
    costs = cluster_distances.sum(axis=1)
    return costs.argmin(axis=0, fill_value=10e9)

distanceMatrix = ssd.squareform(ssd.pdist(features_list, 'cosine'))
clusters = 6
clusters, medoids = cluster(distanceMatrix,clusters)
print(clusters)
print(medoids)

## heatmap of all users
x_labels = ['age','gender','employment','pain_current','rmdq']
#y_labels = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart', 'td', 'tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07', 'tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tsf_tone_10']
y_labels = [str('u'+ str(i)) for i in range(usersNum)]
df_cm = pd.DataFrame(baseline_matrix, y_labels, x_labels)
plt.figure(figsize = (10,10))
plt.title("heatmap of all users ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
sns.heatmap(df_cm, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=0,vmax=4)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()

## add the clustering result to the data
df_cm['cluster'] = clusters

## heatmap for each cluster
df = df_cm.loc[df_cm['cluster'] == medoids[5]]
y_labels = df.index.values
plt.figure(figsize = (10,10))
plt.title("heatmap for cluster 2 ("+str(datetime.now().date())+")")
sns.set(font_scale=1.4)#for label size
df1 = df.drop(['cluster'],axis=1)
sns.heatmap(df1, annot=True, annot_kws={"size": 16}, cmap="YlGnBu", fmt='g',vmin=0,vmax=4)# font size
plt.xlabel('feature')
plt.ylabel('user')
plt.show()
