import numpy as np
import pandas as pd 
import json
import requests
from pandas.io.json import json_normalize
from collections import Counter
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
from pandas.plotting import scatter_matrix
from datetime import datetime
import skfuzzy as fuzz
import time
from time import mktime
from datetime import timedelta
import matplotlib.patches as mpatches

## get backup sessions data
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "sessions" + ".txt"
sessionList = []
with open(path+filename,"r") as file:
    session = eval(file.readline())
    sessionList.extend(session)
    
## normalize time spent
flow_df = pd.DataFrame(sessionList[0])
time_df = pd.DataFrame(sessionList[1])
flow_pca2_norm = (flow_df - flow_df.mean()) / (flow_df.max() - flow_df.min())
time_pca_norm = (time_df - time_df.mean()) / (time_df.max() - time_df.min())
features_list = []
features_list.append(list(flow_pca2_norm[0].values))
features_list.append(list(time_pca_norm[0].values))
features_list = np.transpose(features_list)

## use fpc values to decide number of clusters
X = np.transpose(features_list)
fpcs = []
cntrs = []
num_centers = [2, 3, 4, 5, 6, 7, 8]
for c in num_centers:
    cntr, u, u0, d, jm, p, fpc = fuzz.cluster.cmeans(X,c,m=2,error=0.005,maxiter=1000)
    # cntr is the cluster centers
    # u is the final fuzzy c-partitioned matrix (probability)
    # u0 is the initial guess of fuzzy c-partitioned matrix
    # d is the final euclidean distance matrix
    # jm is the objective function history
    # p is the number of iterations run
    # fpc is fuzzy partition coefficient
    cntrs.append(cntr)
    fpcs.append(fpc)
    cluster_membership = np.argmax(u, axis=0)  # Hardening for visualization
print(fpcs)
plt.figure()
plt.plot(num_centers,fpcs)
plt.xticks(np.arange(min(num_centers),max(num_centers)+1,step=1))
plt.xlabel("Number of centers")
plt.ylabel("Fuzzy partition coefficient")
plt.title("Sessions data FPC result")
plt.grid()
plt.show()

## apply fuzzy cmeans
cntr, u, u0, d, jm, p, fpc = fuzz.cluster.cmeans(X,c=5,m=2,error=0.005,maxiter=1000)  # c is the number of clusters
clusters = np.argmax(u, axis=0)
print(clusters)

## plot clustering result
colors = []
for i in range(len(clusters)):
    if(clusters[i]==0):
        color = 'purple'
    elif(clusters[i]==1):
        color = 'deepskyblue'
    elif(clusters[i]==2):
        color = 'green'
    elif(clusters[i]==3):
        color = 'red'
    elif(clusters[i]==4):
        color = 'magenta'
    elif(clusters[i]==5):
        color = 'blue'
    elif(clusters[i]==6):
        color = 'grey'
    elif(clusters[i]==7):
        color = 'gold'
    colors.append(color)

plt.title("Sessions data Fuzzy C-means clustering result ("+str(datetime.now().date())+")")
plt.xlabel('userflow')
plt.ylabel('timespent')
patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3,patch4],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend
plt.scatter(features_list[:, 0], features_list[:, 1], c=colors)

## get total visits of each user
path = "C:/Users/heyu/PycharmProjects/backup/sessions records/"
filename = "totalVisits" + ".txt"
user_totalVisits = []
with open(path+filename,"r") as file:
    totalVisits = eval(file.readline())
    user_totalVisits.extend(totalVisits)
print(user_totalVisits)

## add clusters ids
df_c = pd.DataFrame(features_list)
df_c['cluster'] = clusters

## get clustering results of each user's sessions
cluster_user = []
starting_index = 0
print(user_totalVisits)
for i in range(len(user_totalVisits)):
    totalVisits = user_totalVisits[i]
    df = df_c.iloc[starting_index:starting_index+totalVisits]
    cluster = list(df['cluster'].values)
    cluster_user.append(cluster)
    starting_index = starting_index + totalVisits
print(len(cluster_user))
print(cluster_user)

## plot user-session-cluster
#users = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
users = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

plt.figure(figsize=(15,10))
plt.title("Users' sessions clustering result ("+str(datetime.now().date())+")")
plt.xlabel('user')
plt.ylabel('session')
plt.yticks(np.arange(0,10,step=1))

patch0 = mpatches.Patch(color='purple',label='cluster 0')
patch1 = mpatches.Patch(color='deepskyblue',label='cluster 1')
patch2 = mpatches.Patch(color='green',label='cluster 2')
patch3 = mpatches.Patch(color='red',label='cluster 3')
patch4 = mpatches.Patch(color='magenta',label='cluster 4')
patch5 = mpatches.Patch(color='blue',label='cluster 5')
patch6 = mpatches.Patch(color='grey',label='cluster 6')
patch7 = mpatches.Patch(color='gold',label='cluster 7')
plt.legend(handles=[patch0,patch1,patch2,patch3,patch4],bbox_to_anchor=(1.05,1),loc=2,borderaxespad=0.)  # set legend

for i in range(len(cluster_user)):
    for j in range(len(cluster_user[i])):
        #plt.scatter(users[i],cluster_user[i][j])
        if(cluster_user[i][j]==0):
            plt.scatter(users[i],j,c='purple')
        elif(cluster_user[i][j]==1):
            plt.scatter(users[i],j,c='deepskyblue')
        elif(cluster_user[i][j]==2):
            plt.scatter(users[i],j,c='green')
        elif(cluster_user[i][j]==3):
            plt.scatter(users[i],j,c='red')
        elif(cluster_user[i][j]==4):
            plt.scatter(users[i],j,c='magenta')
        elif(cluster_user[i][j]==5):
            plt.scatter(users[i],j,c='blue')
            
## show each user's cluster distribution
df_u = pd.DataFrame(cluster_user)

# show the representative cluster of each user
mode_user = []
for i in range(len(cluster_user)):
    data = Counter(cluster_user[i])
    mode = data.most_common(1)
    mode_user.append(mode)
df_u['mode'] = pd.DataFrame(mode_user)

df_u['user'] = ['Tom_TR_Test', 'agnar', 'amar', 'annelovise_no', 'charlotte', 'ilya', 'karen', 'kay', 'keba_en', 'kerstin', 'louise', 'malene', 'nirmalie', 'paul_mail_test_no', 'sadiq', 'stewart','tfs_aud_02', 'tfs_berit_05', 'tfs_jan_01', 'tfs_jorgen_07','tfs_karen_04', 'tfs_lill_03', 'tfs_line_08', 'tfs_oystein_09', 'tfs_steinar_06', 'tsf_tone_10']
#df_u['user'] = [str('u'+ str(i)) for i in range(len(user_totalVisits))]

df_u 
